""" Skywater modelcard

Author: Mario Krattenmacher | Mario.Krattenmacher@semimod.de
"""
# DMT
# Copyright (C) from 2022  SemiMod
# <https://gitlab.com/dmt-development/dmt-device>
#
# This file is part of DMT_core.
#
# DMT_other is free software for non-commercial use only. DMT_other is licensed
# under the DMT License.
#
# DMT_other is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# DMT_LICENSE.md in the root directory of DMT for more details.
#
# You should have received a copy of the DMT License along with this program.
from __future__ import annotations
import copy

try:
    from semver.version import Version as VersionInfo
except ImportError:
    from semver import VersionInfo

from DMT.core import MCard, McParameter, unit_registry
from DMT.core.circuit import SGP_BJT, Circuit, CircuitElement, RESISTANCE, VOLTAGE

SEMVER_MCSKYWATER_CURRENT = VersionInfo(major=1, minor=0)


class McSkywater(MCard):
    """All model parameters of Skywater130 pdk models

    Parameters
    -----------
    load_model_from_path : str, optional
        Initialise the modelcard with the parameter from the given file path.
    version : float, optional
        Version of the model card. Default is 1.0
    pdk_path : str
        Path to the Skywater pdk: ".../sky130.lib.spice"
    pdk_corner : str
        Corner for this modelcard ("tt" ...)
    """

    def __init__(
        self,
        load_model_from_path=None,
        version=1.0,
        default_circuit="common_source",
        __McSkywater__=SEMVER_MCSKYWATER_CURRENT,
        nodes_list=("D", "G", "S", "B"),
        default_subckt_name="X1",
        default_module_name="sky130_fd_pr__nfet",
        possible_groups=None,
        vae_module=None,
        **kwargs,
    ):
        if possible_groups is None:
            possible_groups = {
                "Geometrie": "geo",
            }

        super().__init__(
            nodes_list,
            default_subckt_name,
            default_module_name,
            version=version,
            possible_groups=possible_groups,
            vae_module=vae_module,
            **kwargs,
        )

        if not isinstance(__McSkywater__, VersionInfo):
            try:
                __McSkywater__ = VersionInfo.parse(__McSkywater__)
            except TypeError:
                __McSkywater__ = VersionInfo.parse(f"{__McSkywater__:1.1f}.0")

        if __McSkywater__ != SEMVER_MCSKYWATER_CURRENT:
            raise IOError("DMT->McSkywater: The given version of __McSkywater__ is unknown!")

        if self.va_codes:
            raise IOError("The Skywater mcard has no VA-Code!")

        self.add(McParameter("l", value=1, unit=unit_registry.metre, group="geo"))
        self.add(McParameter("w", value=1, unit=unit_registry.metre, group="geo"))
        self.add(McParameter("nf", value=1, unit=unit_registry.dimensionless, group="geo"))
        self.add(McParameter("ad", value=0, unit=unit_registry.dimensionless, group="geo"))
        self.add(McParameter("as", value=0, unit=unit_registry.dimensionless, group="geo"))
        self.add(McParameter("pd", value=0, unit=unit_registry.dimensionless, group="geo"))
        self.add(McParameter("ps", value=0, unit=unit_registry.dimensionless, group="geo"))
        self.add(McParameter("nrd", value=0, unit=unit_registry.dimensionless, group="geo"))
        self.add(McParameter("nrs", value=0, unit=unit_registry.dimensionless, group="geo"))
        self.add(McParameter("sa", value=0, unit=unit_registry.dimensionless, group="geo"))
        self.add(McParameter("sb", value=0, unit=unit_registry.dimensionless, group="geo"))
        self.add(McParameter("sd", value=0, unit=unit_registry.dimensionless, group="geo"))
        self.add(McParameter("mult", value=1, unit=unit_registry.dimensionless, group="geo"))

        if load_model_from_path is not None:
            super().load_model_parameters(
                load_model_from_path, force=False
            )  # False: there may be invalid parameters in the file

        self.default_circuit = default_circuit

    def info_json(self, **kwargs):
        """Returns a dict with serializeable content for the json file to create. Add the info about the concrete subclass to create here!"""
        info_dict = super(McSkywater, self).info_json(**kwargs)
        info_dict["__McSkywater__"] = str(
            SEMVER_MCSKYWATER_CURRENT
        )  # make versions, so we can introduce compatibility here!
        info_dict["default_circuit"] = self.default_circuit

        return info_dict

    def get_circuit(self, use_build_in=False, topology=None, **kwargs) -> Circuit:
        """Here the modelcard defines it's default simulation circuit.

        Parameters
        -----------
        use_build_in : {False, True}, optional
            Creates a circuit for the modelcard using the build-in model
        topology : optional
            If a model has multiple standard circuits, use the topology to differentiate between them..
        """
        if use_build_in:
            mcard = self.get_build_in()
        else:
            mcard = self

        if topology is None:
            topology = self.default_circuit

        if topology == "common_source":
            return Circuit(
                [
                    CircuitElement(
                        self.default_module_name,
                        "x1",
                        ["n_D", "n_G", "n_S", "n_B"],
                        parameters=mcard,
                    ),
                    CircuitElement(RESISTANCE, "R_D", ["n_D", "n_DX"], parameters=[("R", "1")]),
                    CircuitElement(
                        VOLTAGE, "V_D", ["n_DX", "0"], parameters=[("Vdc", "V_D"), ("Vac", "1")]
                    ),
                    CircuitElement(
                        VOLTAGE, "V_G", ["n_GX", "0"], parameters=[("Vdc", "V_G"), ("Vac", "1")]
                    ),
                    CircuitElement(RESISTANCE, "R_G", ["n_GX", "n_G"], parameters=[("R", "1")]),
                    CircuitElement(RESISTANCE, "R_S", ["n_S", "0"], parameters=[("R", "0")]),
                    CircuitElement(RESISTANCE, "R_B", ["n_B", "0"], parameters=[("R", "0")]),
                ]
            )
        else:
            raise NotImplementedError(f"DMT.McSkywater: unknown topology {topology}")

    def get_build_in(self):
        """Return the parameters embedded in a build-in model (no Va code and correct module name etc)"""
        return copy.deepcopy(self)

    def get_clean_modelcard(self):
        """Returns all parameters which are part of SGP and adds correct units"""
        default_mcard = McSkywater(
            version=self.version,
            default_circuit=self.default_circuit,
            pdk_path=self.pdk_path,
            pdk_corner=self.pdk_corner,
        )
        for para in self:
            if para in default_mcard:
                default_mcard.set(para)
            else:
                print("Warning: parameter " + para.name + " was removed by clean_mcard.")

        return default_mcard



""" Database management

Class which represents all common features in DMT that are used to search, organize and filter data.
Features:

* Can read all relevant file formats produced by simulators used at CEDIC.
* Uses pandas to manage the data internally.
* Allows to "search" and "filter" existing data.

"""
# DMT_core
# Copyright (C) from 2022  SemiMod
# Copyright (C) until 2021  Markus Müller, Mario Krattenmacher and Pascal Kuthe
# <https://gitlab.com/dmt-development/dmt-core>
#
# This file is part of DMT_core.
#
# DMT_core is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DMT_core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
import pandas as pd
import warnings
import _pickle as cpickle
from pathlib import Path

from DMT.config import DATA_CONFIG
from DMT.core.data_frame import DataFrame
from DMT.core.singleton import Singleton
from DMT.core.naming import SpecifierStr


class DatabaseManager(object, metaclass=Singleton):
    """Class responsible for data management in DMT.

    Methods
    -------
    load_db(db_dir):
        Load a full db from the given path.

    save_db(db_dir, data):
        Save a full db to the given path.

    del_db(db_dir)
        Delete the hole database.

    save_df(df, file_name):
        Save a dataframe as file_name, where file_name is the direct path to the desired file.

    load_df(file_name):
        Loads a DMT dataframe from the file.
    """

    def load_db(self, db_dir):
        """Loads all DMT.DataFrames from the database at db_dir using implicit casting.

        Parameters
        ----------
        db_dir  : str or os.Pathlike
            Path to the database that shall be used to load df.

        Raises
        ------


        Returns
        -------
        data : {key: DataFrame}
            Loaded dataframes
        """
        if not isinstance(db_dir, Path):
            db_dir = Path(db_dir)

        if not db_dir.exists():
            raise FileNotFoundError

        data = {}

        if DATA_CONFIG["useHDF5Store"]:
            try:
                db = pd.io.pytables.HDFStore(
                    str(db_dir.expanduser()), mode="r", complevel=9, complib="zlib"
                )

                keys_db = db.keys()  # sorry :(
                for key in keys_db:
                    df = db.get(key)
                    df.__class__ = DataFrame  # cast to DMT.DataFrame
                    key = key[1:]  # remove starting '/'

                    # cast back
                    dict_reconvert = {}
                    for col in df.columns:
                        dict_reconvert[col] = SpecifierStr.string_from_load(col)

                    df.rename(columns=dict_reconvert, inplace=True)
                    # prevent invisible column bug:
                    df = df.loc[:, ~df.columns.duplicated()]
                    if not df.columns.is_unique:
                        raise IOError

                    data[key] = df
                db.close()
            except (OSError, RuntimeError):
                with db_dir.open(mode="rb") as my_db:
                    data = cpickle.load(my_db)

        else:
            with db_dir.open(mode="rb") as my_db:
                data = cpickle.load(my_db)

        # cast back
        for key, df in data.items():
            if isinstance(df, pd.DataFrame):
                df.__class__ = DataFrame

        return data

    def save_db(self, db_dir, data):
        """Save the complete data into the database of dut. The old database is overwritten!

        Parameters
        ----------
        db_dir   :  str or os.Pathlike
            Path to the database that shall be used to save the data.
        data     :  {key: DataFrame}
            Dataframes to save
        """
        if not isinstance(db_dir, Path):
            db_dir = Path(db_dir)

        # create all directories until database
        db_dir.parent.mkdir(parents=True, exist_ok=True)

        # convert to usual pandas dataframe
        try:
            for key, df in data.items():
                if isinstance(df, DataFrame):
                    df.__class__ = pd.DataFrame
                elif isinstance(df, pd.DataFrame):
                    pass
                else:
                    raise TypeError("Only DMT.core.DataFrame can be converted to pandas.DataFrame!")
        except TypeError:
            # if it is not a DataFrame, just directly dump it.
            with db_dir.open(mode="wb") as my_db:
                cpickle.dump(data, my_db)

            return

        if DATA_CONFIG["useHDF5Store"]:
            db = pd.io.pytables.HDFStore(
                str(db_dir.expanduser()), mode="w", complevel=9, complib="zlib"
            )  # file is created if it does not exist
            for key, df in data.items():
                # cast to unique string
                dict_convert = {}
                dict_reconvert = {}
                for col in df.columns:
                    try:
                        dict_convert[col] = col.string_to_save()
                        dict_reconvert[col.string_to_save()] = col
                    except AttributeError:
                        pass
                df = df.rename(columns=dict_convert)

                with warnings.catch_warnings():
                    warnings.simplefilter("ignore")
                    db.put(key, df)

                # cast back
                # for col in df.columns:
                #     df = df.rename(columns={SpecifierStr.string_from_load(col):col})

                df = df.rename(columns=dict_reconvert)
            db.close()
        else:
            with db_dir.open(mode="wb") as my_db:
                cpickle.dump(data, my_db)

        # convert back
        for key, df in data.items():
            df.__class__ = DataFrame

    def del_db(self, db_dir):
        """Delete the .h5 database for dut.

        Parameters
        ----------
        db_dir   :  string or os.Pathlike
            Full path to the database.
        """
        if not isinstance(db_dir, Path):
            db_dir = Path(db_dir)

        if db_dir.exists():
            db_dir.unlink()

    def save_df(self, df, file_name):
        """Save the data stored in df as file_name, where file_name is the direct path to the file.

        Parameters
        ----------
        df  :  DMT.core.DataFrame
            A dataframe object that shall be saved.
        file_name  :  str or os.Pathlike
            Direct path to the file
        """
        if not isinstance(file_name, Path):
            file_name = Path(file_name)

        file_name.parent.mkdir(parents=True, exist_ok=True)

        df.to_pickle(str(file_name))

    def load_df(self, file_name):
        """Load the data stored in file_name, where file_name is the direct path to the file.

        Parameters
        ----------
        file_name  :  str
            Direct path to the file

        Returns
        -------
        df  :  DMT.core.DataFrame
            Loaded dataframe object.
        """
        df = pd.read_pickle(str(file_name))
        df.__class__ = DataFrame
        return df



r""" Global namings for DMT

Internally all variables and column names which contain a quantity of a specifier must have the same given names, e.g. all voltages will be called: 'V\_'
"""
# DMT_core
# Copyright (C) from 2022  SemiMod
# Copyright (C) until 2021  Markus Müller, Mario Krattenmacher and Pascal Kuthe
# <https://gitlab.com/dmt-development/dmt-core>
#
# This file is part of DMT_core.
#
# DMT_core is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DMT_core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>

from __future__ import annotations
import copy
import numpy as np
import warnings
from typing import List, Union, Set, FrozenSet
from pint.formatting import siunitx_format_unit
from more_itertools import unique_everseen
from DMT.config import DATA_CONFIG
from DMT.core import Singleton, unit_registry
import pint

UNIT_PREFIX = {
    1e-12: r"\tera",
    1e-9: r"\giga",
    1e-6: r"\mega",
    1e-3: r"\kilo",
    1: r"",
    1e3: r"\milli",
    1e6: r"\micro",
    1e9: r"\nano",
    1e12: r"\pico",
    1e15: r"\femto",
}
UNIT_PREFIX_MIX = {
    "J": {  # current density-> they are all over the place :/
        1e-9: r"\milli\ampere\per\square\micro\meter",
        1e-6: r"\micro\ampere\per\square\micro\meter",
        1e-3: r"\milli\ampere\per\square\micro\meter",
        1e-0: r"\ampere\per\square\micro\meter",
        1e+3: r"\milli\ampere\per\square\micro\meter",
    },
    "F": {
        1: r"\volt\per\meter",
        1e-5: r"\kilo\volt\per\centi\meter",
    },  # field
}
UNIT_PREFIX_DENOMINATOR = {
    1e-6: r"\centi",
}


class SpecifierStr(str):
    """Acts like a string, but at the same time has the attribute "nodes"

    https://stackoverflow.com/a/2673863

    Yeah it is already implemented :P

    Parameters
    ----------
    specifier : str
    nodes : [str], optional
    sub_specifiers : [str], str, {'', sub_specifier}, optional

    Attributes
    ----------
    specifier : str
    nodes : [str]
    sub_specifiers : [str]

    """

    specifier: str = ""
    nodes: List[str] = []
    sub_specifiers: FrozenSet[str] = []

    # pylint: disable=no-member
    def __new__(
        cls,
        specifier: Union[str, SpecifierStr],
        *nodes: str,
        sub_specifiers: Union[
            FrozenSet[Union[str, SpecifierStr]],
            Set[Union[str, SpecifierStr]],
            str,
            SpecifierStr,
            None,
        ] = None,
    ):
        if not nodes and sub_specifiers is None:
            try:
                nodes = specifier.nodes  # type: ignore
                sub_specifiers = specifier.sub_specifiers  # type: ignore
                specifier = specifier.specifier  # type: ignore
            except AttributeError:
                # revert old state if an Attribute error occurs!
                nodes = ()
                sub_specifiers = None

        # cast to set
        if sub_specifiers is None:
            sub_specifiers = frozenset()
        elif isinstance(sub_specifiers, str):
            sub_specifiers = frozenset([sub_specifiers])

        tempset = set()
        for sub_specifier in sub_specifiers:
            if sub_specifier:
                if isinstance(sub_specifier, SpecifierStr):
                    for sub_spec in sub_specifier.sub_specifiers:
                        tempset.add(sub_spec)
                else:
                    tempset.add(sub_specifier)
        sub_specifiers = frozenset(str(sub_spec) for sub_spec in tempset)  # string cast :/

        # check if all are valid sub_specifiers
        # if any(sub_specifier not in SUB_SPECIFIERS_STR for sub_specifier in sub_specifiers):
        #     # add the new sub_specifier to naming.sub_specifiers and also to SUB_SPECIFIERS_STR
        #     # Not done because parameter sub_specifiers replaces the module attribute
        #     raise SpecifierNotKnown('The sub_specifiers part of a column name can only be one of the following:\n' + str(SUB_SPECIFIERS_STR))

        if nodes and "".join(nodes):  # string cast :/
            nodes = [str(node) for node in nodes]  # type: ignore
            node_str = "_" + "".join(nodes)
        else:
            node_str = ""
            nodes = []  # type: ignore
            # make sure nodes is a list not a tuple

        if sub_specifiers:
            sub_specifiers_str = "|" + "|".join(
                sorted(sub_specifiers)
            )  # sort the sub specifiers to ensure same string always!
        else:
            sub_specifiers_str = ""

        col_name = str(specifier) + node_str + sub_specifiers_str

        obj = str.__new__(cls, col_name)
        obj.specifier = str(specifier)
        obj.nodes = nodes  # type: ignore
        obj.sub_specifiers = sub_specifiers  # type: ignore
        return obj

    def get_tex_unit(self, scale=1, add="") -> str:
        r"""Get the unit of a given specifier in base units.

        If scale is different from one, a suitable unit prefix is chosen. E.g. scale=1e3 and specifier=CURRENT -> \si{\milli\ampere}

        Parameters
        ----------
        scale : float, integer
            Unit prefix determination scale.
        add : string
            Additional string that is added to the unit

        Returns
        -------
        unit : string
            TeX representation of the specifer's unit
        """
        unit = self.get_pint_unit()

        if sub_specifiers.PHASE.sub_specifiers <= self.sub_specifiers:
            return r"\si{\degree}"

        elif self.specifier in UNIT_PREFIX_MIX:  # mixed unit
            try:
                unit_prefix_mix = UNIT_PREFIX_MIX[self.specifier][np.round(scale, decimals=10)]
                return r"\si{" + unit_prefix_mix + r"}"
            except KeyError as err:
                raise IOError(
                    "DMT -> Plot: natural unit for " + self.specifier + " not implemented."
                ) from err

        else:  # pure unit
            # for non-mixed quantities like voltages and current

            # get unit from pint
            try:
                unit = siunitx_format_unit(unit)  # type: ignore
            except TypeError:
                unit = siunitx_format_unit(
                    unit._units, unit_registry
                )  # new version has other interface

            if unit.startswith("\\per"):
                # for 1/m^3 -> unit prefix should be in the denominator
                unit = "\\per" + UNIT_PREFIX_DENOMINATOR[scale] + unit[4:]
            else:
                unit = UNIT_PREFIX[scale] + unit

            return r"\si{" + unit + add + r"}"

    def to_label(self, scale=1, negative=False, divide_by_unit=False, **kwargs) -> str:
        """Generates a label for plots for this specifier, where scale determines the unit prefix.

        Parameters
        ----------
        scale : float, integer, optional
            Unit prefix determination scale.

        negative : bool, optional
            If True, a minus sign is added before the label

        divide_by_unit : bool, optional
            If True, the unit is given as division, if False in brackets.

        kwargs : optional
            Keyword arguments passed on to self.to_tex(). Currently this allows subscript and superscript strings.

        Returns
        -------
        unit : string
            TeX representation of the specifer's unit
        """
        tex_unit = self.get_tex_unit(scale=scale)
        if divide_by_unit:
            if tex_unit == unit_registry.dimensionless or tex_unit == "\\si{}":
                tex_unit = ""
            else:
                tex_unit = "/" + tex_unit
            if negative:
                return r"$-" + self.to_tex(**kwargs) + tex_unit + r"$"
            else:
                return r"$" + self.to_tex(**kwargs) + tex_unit + r"$"
        else:
            left_brack = r"\left("
            right_brack = r"\right)"
            if tex_unit == unit_registry.dimensionless or tex_unit == "\\si{}":
                left_brack = ""
                right_brack = ""
            if negative:
                return r"$-" + self.to_tex(**kwargs) + left_brack + tex_unit + right_brack + r"$"
            else:
                return r"$" + self.to_tex(**kwargs) + left_brack + tex_unit + right_brack + r"$"

    def to_legend_with_value(self, value, scale=1, decimals=2, **kwargs) -> str:
        """Creates a SI legend entry in the form : specifier_tex = \\SI{value}{scale, spec_unit}

        Parameters
        ----------
        value : float
            Value to print into the legend
        scale : int, optional
            Scaling of unit and value using SI prefixes, by default 1.
        **kwargs
            Are passed to specifier.to_tex

        Returns
        -------
        str
            math tex legend entry
        """
        value = np.real(value)

        unit = self.get_pint_unit()
        try:
            unit = siunitx_format_unit(unit)  # type: ignore
        except TypeError:
            unit = siunitx_format_unit(
                unit._units, unit_registry
            )  # new version has other interface
        # for non-mixed quantities like voltages and current
        if self.specifier in UNIT_PREFIX_MIX:  # mixed unit
            unit_with_prefix = UNIT_PREFIX_MIX[self.specifier][np.round(scale, decimals=10)]

            return f"${self.to_tex(**kwargs):s}=\\SI{{{value * scale:.{decimals}f}}}{{{unit_with_prefix:s}{unit:s}}}$"

        else:
            unit_prefix = UNIT_PREFIX[scale]

            return f"${self.to_tex(**kwargs):s}=\\SI{{{value * scale:.{decimals}f}}}{{{unit_prefix:s}{unit:s}}}$"

    def to_raw_string(self) -> str:
        """get a raw string from the specifier -> can be used for string operations and variable naming...

        Returns
        -------
        str : string
            string conversion of self...
        """
        return str(self)  # not tested

    def string_to_save(self) -> str:
        """Generates a single string which allows to identify the specifers, nodes and subspecifiers. Used to save into HDF5 as there the columns have to be valid strings."""
        return (
            "$"
            + str(self.specifier)
            + "#"
            + "?".join([str(node) for node in self.nodes])
            + "!"
            + "|".join([str(sub_specifier) for sub_specifier in self.sub_specifiers])
        )

    @classmethod
    def string_from_load(cls, string) -> str:
        """Generates a SpecifierStr from a string if the string was generated using :meth:`~DMT.core.specifiers.SpecifierStr.string_to_save`."""
        if string.startswith("$"):  # SpecifierStr
            splitted = string[1:].split("#")
            spec = splitted[0]
            splitted = splitted[1].split("!")

            nodes = splitted[0].split("?")
            sub_specifiers = splitted[1].split("|")
            return SpecifierStr(spec, *nodes, sub_specifiers=sub_specifiers)
        else:  # it was just a regular str
            return string

    def __add__(
        self: SpecifierStr, other: Union[SpecifierStr, str, list[Union[str, SpecifierStr]]]
    ) -> SpecifierStr:
        """Method stub"""
        raise NotImplementedError

    def __radd__(self, other):
        # if isinstance(other, SpecifierStr): # not needed since __add__ is taking care of this!
        if isinstance(other, str):  # reflected str + SpecifierStr
            return SpecifierStr(
                other + self.specifier, *self.nodes, sub_specifiers=self.sub_specifiers
            )
        else:
            return NotImplemented

    def __eq__(self, other: Union[str, SpecifierStr]) -> bool:
        """Checks if two objects are equal

        Parameters
        ----------
        other : str | SpecifierStr
            Compare to this column name

        Returns
        -------
        bool
            True if equal
        """
        if isinstance(other, SpecifierStr):
            if self.specifier == other.specifier and self.nodes == other.nodes:
                # order of subspecifiers does not matter Oo -> can not since these are sets
                return self.sub_specifiers == other.sub_specifiers
        elif isinstance(other, str):
            if str(self) == other:
                return True
            else:
                col_other = get_specifier_from_string(other, nodes=self.nodes)
                if isinstance(col_other, SpecifierStr):  # here also the order does not matter -.-
                    return self == col_other
        elif other is not None:
            return NotImplemented

        return False

    def __contains__(self, other: Union[str, SpecifierStr]) -> bool:
        """String contains

        Parameters
        ----------
        other : str | SpecifierStr
            Other string or specifier which may contains self

        Returns
        -------
        bool
            str(other) in str(self)
        """
        if isinstance(other, SpecifierStr):
            if other.specifier and not other.nodes and not other.sub_specifiers:
                if self.specifier == other.specifier:
                    return True
            elif not other.specifier and other.nodes and not other.sub_specifiers:
                if all(node in self.nodes for node in other.nodes):
                    return True
            elif not other.specifier and not other.nodes and other.sub_specifiers:
                if other.sub_specifiers <= self.sub_specifiers:
                    return True

            return False

        elif isinstance(other, str):
            if str(self) == other:
                return True
            else:
                col_other = get_specifier_from_string(other, nodes=self.nodes)
                if isinstance(col_other, SpecifierStr):
                    return col_other in self

        elif other is not None:
            return NotImplemented

        # return str(other) in str(self)
        return False

    def __hash__(self):
        return hash(str(self))

    def get_pint_unit(self) -> pint.Unit:
        """Declaration of function, to be set later

        Returns
        -------
        pint.UnitRegistry
            [description]

        Raises
        ------
        NotImplementedError
            [description]
        """
        raise NotImplementedError()

    def get_descriptor(self) -> str:
        """Declaration of function, to be set later

        Returns
        -------
        str
            [description]

        Raises
        ------
        NotImplementedError
            [description]
        """
        raise NotImplementedError

    def to_tex(self) -> str:
        """Declaration of function, to be set later

        Returns
        -------
        str
            [description]

        Raises
        ------
        NotImplementedError
            [description]
        """
        raise NotImplementedError


class GlobalObj(object):
    """A simple class used to store class variables that are global. Also one can iterate over the class attributes in a sorted way."""

    _MEMBERS = []

    def __iter__(self):
        if not self._MEMBERS:
            self._set_members()

        for elem in list.__iter__(self._MEMBERS):
            yield elem

    def _set_members(self):
        """sets self._MEMBERS in a sorted fashion with a possible additional list."""
        members = [
            getattr(self, attribute)
            for attribute in dir(self)
            if not attribute.startswith("__")
            and not callable(getattr(self, attribute))
            and not attribute == "_MEMBERS"
        ]
        self._MEMBERS = sorted(members, key=len, reverse=True)


#############################################
# HOLY DEFINITION OF DMT NAMING CONVENTIONS #
#############################################
class _sub_specifiers(GlobalObj, metaclass=Singleton):

    PERIMETER = SpecifierStr("", sub_specifiers="PERI")
    AREA = SpecifierStr("", sub_specifiers="AREA")
    LENGTH = SpecifierStr("", sub_specifiers="LENGTH")
    WIDTH = SpecifierStr("", sub_specifiers="WIDTH")
    CORNER = SpecifierStr("", sub_specifiers="CORNER")
    FORCED = SpecifierStr("", sub_specifiers="FORCED")
    AC_BASE = SpecifierStr("", sub_specifiers="AC_BASE")
    AC_COLLECTOR = SpecifierStr("", sub_specifiers="AC_COLLECTOR")
    AC_GATE = SpecifierStr("", sub_specifiers="AC_GATE")
    AC_DRAIN = SpecifierStr("", sub_specifiers="AC_DRAIN")
    AC = SpecifierStr("", sub_specifiers="AC")
    REAL = SpecifierStr("", sub_specifiers="REAL")
    IMAG = SpecifierStr("", sub_specifiers="IMAG")
    MAG = SpecifierStr("", sub_specifiers="MAG")
    PHASE = SpecifierStr("", sub_specifiers="PHASE")
    NOISE = SpecifierStr("", sub_specifiers="NOISE")
    QUASISTATIC = SpecifierStr("", sub_specifiers="QUASISTATIC")
    JUNCTION = SpecifierStr("", sub_specifiers="JUNCTION")
    MINORITY = SpecifierStr("", sub_specifiers="MINORITY")
    MAJORITY = SpecifierStr("", sub_specifiers="MAJORITY")
    ELECTRONS = SpecifierStr("", sub_specifiers="ELECTRONS")
    HOLES = SpecifierStr("", sub_specifiers="HOLES")
    CHANNEL = SpecifierStr("", sub_specifiers="CHANNEL")
    MAX = SpecifierStr("", sub_specifiers="MAX")
    MIN = SpecifierStr("", sub_specifiers="MIN")
    MID = SpecifierStr("", sub_specifiers="MID")
    TRAPS = SpecifierStr("", sub_specifiers="TRAPS")
    YDIR = SpecifierStr("", sub_specifiers="YDIR")
    XDIR = SpecifierStr("", sub_specifiers="XDIR")
    ZDIR = SpecifierStr("", sub_specifiers="ZDIR")

    def add_members(self, members):
        for name, value in members:
            if name in dir(self):
                raise OSError("The subspecifier " + name + " already exists!")
            setattr(self, name, SpecifierStr("", sub_specifiers=value))

        self._set_members()


class _specifiers(GlobalObj, metaclass=Singleton):
    # Only Natural Quantities here
    VOLTAGE = SpecifierStr("V")
    FIELD = SpecifierStr("F")
    VELOCITY = SpecifierStr("VELO")
    GRADING = SpecifierStr("GRADING")
    RESISTANCE = SpecifierStr("R")
    POWER = SpecifierStr("P")
    CURRENT = SpecifierStr("I")
    CAPACITANCE = SpecifierStr("C")
    CHARGE = SpecifierStr("Q")
    INDUCTANCE = SpecifierStr("L")
    TEMPERATURE = SpecifierStr("TEMP")
    TIME = SpecifierStr("TIME")
    FREQUENCY = SpecifierStr("FREQ")
    VOLTAGE = SpecifierStr("V")
    ELECTRONS = SpecifierStr("N")
    CONDUCTION_BAND_EDGE = SpecifierStr("EC")
    VALENCE_BAND_EDGE = SpecifierStr("EV")
    HOLES = SpecifierStr("P")
    NET_DOPING = SpecifierStr("NNET")
    DONNORS = SpecifierStr("DON")
    ACCEPTORS = SpecifierStr("ACC")
    CURRENT_DENSITY = SpecifierStr("J")
    SPACE_CHARGE = SpecifierStr("RHO")
    QUASI_FERMI_POTENTIAL = SpecifierStr("PHI")
    ELECTRICAL_POTENTIAL = SpecifierStr("PSI")
    MOBILITY = SpecifierStr("MU")
    X = SpecifierStr("x")
    Y = SpecifierStr("y")
    Z = SpecifierStr("z")
    ENERGY = SpecifierStr("E")

    # only derived quantities here
    DC_CURRENT_AMPLIFICATION = SpecifierStr("BETA")
    TRANSCONDUCTANCE = SpecifierStr("GM")
    TRANSIT_FREQUENCY = SpecifierStr("F_T")
    MAXIMUM_OSCILLATION_FREQUENCY = SpecifierStr("F_MAX")
    MAXIMUM_AVAILABLE_GAIN = SpecifierStr("MAG")
    MAXIMUM_STABLE_GAIN = SpecifierStr("MSG")
    TRANSIT_TIME = SpecifierStr("TAU_F")
    UNILATERAL_GAIN = SpecifierStr("GU")

    # SpecifierStr objects for small signal parameters
    SS_PARA_S = SpecifierStr("S")
    SS_PARA_H = SpecifierStr("H")
    SS_PARA_Y = SpecifierStr("Y")
    SS_PARA_Z = SpecifierStr("Z")
    SS_PARA_A = SpecifierStr("A")
    SS_PARA_T = SpecifierStr("T")

    def add_members(self, members):
        for name, value in members:
            if name in dir(self):
                raise OSError("The specifier " + name + " already exists!")
            setattr(self, name, SpecifierStr(value))

        self._set_members()


class _specifiers_ss_para(GlobalObj, metaclass=Singleton):
    # SpecifierStr objects for small signal parameters, need to be repeated since these specifiers have unique properties
    SS_PARA_S = SpecifierStr("S")
    SS_PARA_H = SpecifierStr("H")
    SS_PARA_Y = SpecifierStr("Y")
    SS_PARA_Z = SpecifierStr("Z")
    SS_PARA_A = SpecifierStr("A")
    SS_PARA_T = SpecifierStr("T")


specifiers_ss_para: _specifiers_ss_para = _specifiers_ss_para()
"""Small signal parameters known to DMT."""

specifiers: _specifiers = _specifiers()
"""Specifiers known to DMT. In a written form these would be the variable."""

sub_specifiers: _sub_specifiers = _sub_specifiers()
"""Sub specifiers known to DMT. In a written form these would be placed in the subscript of a variable."""

specifiers.add_members(DATA_CONFIG["custom_specifiers"])
sub_specifiers.add_members(DATA_CONFIG["custom_sub_specifiers"])

# needed for addition
SUB_SPECIFIERS_STR = [
    str(member).replace("|", "") for member in sub_specifiers._MEMBERS
]  # pylint: disable=protected-access

##################################################
# add methods to SpecifierStr that depend on the defined specifiers above
##################################################


def add(self: SpecifierStr, other: Union[SpecifierStr, str, List[Union[str, SpecifierStr]]]):
    """Method is defined later, since we need the SUB_SPECIFIERS_STR list here...thanks python"""
    if isinstance(other, SpecifierStr):
        if (self.specifier == other.specifier) or (not other.specifier):
            spec_new = self.specifier
        elif not self.specifier:
            spec_new = other.specifier
        else:
            raise IOError(
                f"DMT->Naming: Adding 2 different SpecifierStr, which specifier to take:\n{self.specifier} or {other.specifier} ?"
            )

        nodes_new = self.nodes + other.nodes
        sub_specifiers_new = self.sub_specifiers | other.sub_specifiers
        return SpecifierStr(spec_new, *nodes_new, sub_specifiers=sub_specifiers_new)

    elif isinstance(other, str):
        if other == "":
            return self

        if other[0] == "_" or other[0] == "|":
            other = other[1:]

        if other in SUB_SPECIFIERS_STR:
            return SpecifierStr(
                self.specifier, *self.nodes, sub_specifiers=self.sub_specifiers | {other}
            )
        else:
            return SpecifierStr(
                self.specifier, *self.nodes, other, sub_specifiers=self.sub_specifiers
            )  # @Mario we should discuss this

    elif isinstance(other, (list, set, frozenset)):
        if self.nodes:
            return SpecifierStr(
                self.specifier, *self.nodes, sub_specifiers=self.sub_specifiers | set(other)  # type: ignore
            )
        else:
            return SpecifierStr(self.specifier, *other, sub_specifiers=self.sub_specifiers)

    else:
        return NotImplemented


SpecifierStr.__add__ = add


def get_pint_unit(self) -> pint.Unit:
    """Return the DMT base unit of this specifier as a pint unit"""
    if sub_specifiers.PHASE.sub_specifiers <= self.sub_specifiers:
        return unit_registry.degree

    unit_converter = {
        specifiers_ss_para.SS_PARA_Y: unit_registry.siemens,
        specifiers_ss_para.SS_PARA_H: unit_registry.dimensionless,
        specifiers_ss_para.SS_PARA_S: unit_registry.dimensionless,
        specifiers.UNILATERAL_GAIN: unit_registry.dimensionless,
        specifiers.VOLTAGE: unit_registry.volt,
        specifiers.CAPACITANCE: unit_registry.farad,
        specifiers.CHARGE: unit_registry.coulomb,
        specifiers.FREQUENCY: unit_registry.hertz,
        specifiers.CURRENT: unit_registry.ampere,
        specifiers.CURRENT_DENSITY: unit_registry.ampere_per_square_meter,
        specifiers.TEMPERATURE: unit_registry.kelvin,
        specifiers.RESISTANCE: unit_registry.ohm,
        specifiers.POWER: unit_registry.watt,
        specifiers.TIME: unit_registry.second,
        specifiers.X: unit_registry.meter,
        specifiers.TRANSIT_FREQUENCY: unit_registry.hertz,
        specifiers.MAXIMUM_OSCILLATION_FREQUENCY: unit_registry.hertz,
        specifiers.TRANSIT_TIME: unit_registry.second,
        specifiers.TRANSCONDUCTANCE: unit_registry.siemens,
        specifiers.ENERGY: unit_registry.volt,
        specifiers.FIELD: unit_registry.volt / unit_registry.meter,
        specifiers.DC_CURRENT_AMPLIFICATION: unit_registry.dimensionless,
        specifiers.MAXIMUM_STABLE_GAIN: unit_registry.dimensionless,
        specifiers.NET_DOPING: 1 / unit_registry.meter / unit_registry.meter / unit_registry.meter,
        specifiers.ACCEPTORS: 1 / unit_registry.meter / unit_registry.meter / unit_registry.meter,
        specifiers.DONNORS: 1 / unit_registry.meter / unit_registry.meter / unit_registry.meter,
    }  # type: dict[SpecifierStr, pint.Unit]
    try:
        return unit_converter[self.specifier]
    except KeyError as err:
        raise IOError(
            "DMT -> Naming: no unit defined for specifier " + self.specifier + "."
        ) from err


SpecifierStr.get_pint_unit = get_pint_unit


def get_descriptor(self):
    """Can be generalized using the _specifiers class."""
    descriptor = {
        specifiers.CURRENT: "current",
        specifiers.CURRENT_DENSITY: "current density",
        specifiers.CAPACITANCE: "capacitance",
        specifiers.VOLTAGE: "voltage",
        specifiers.RESISTANCE: "resistance",
        specifiers.POWER: "power",
        specifiers.ENERGY: "energy",
    }
    return descriptor[self.specifier]


SpecifierStr.get_descriptor = get_descriptor


def to_tex(self, subscript="", superscript=""):
    """Return a Tex representation of this specifier. If subscript is not None the string in subscript will be appended. If superscript is not None, a superscript will be added.

    Parameters
    ----------
    subscript : str
        Subscript to add to the specifier's Tex representation
    superscript : str
        Superscript to add to the specifier's Tex representation

    Returns
    -------
    tex : str
            A Tex representation fo the specifier.
    """
    if self.specifier in specifiers_ss_para:
        nodes_temp = copy.deepcopy(self.nodes)

        for i_nodes, node in enumerate(nodes_temp):
            node = node.replace("B", "1")
            node = node.replace("C", "2")
            node = node.replace("G", "1")
            node = node.replace("D", "2")
            nodes_temp[i_nodes] = node

        if subscript == "":
            tex = (
                r"\underline{"
                + str(self.specifier)
                + r"}_{\mathrm{"
                + str("".join(nodes_temp))
                + r"}}"
            )
        else:
            tex = (
                r"\underline{"
                + str(self.specifier)
                + r"}_{\mathrm{"
                + str("".join(nodes_temp))
                + r","
                + subscript
                + r"}}"
            )

    # first catch special cases and add subscript
    elif self.specifier == specifiers.TRANSIT_FREQUENCY:
        tex = r"f_{\mathrm{T" + subscript + r"}}"
    elif self.specifier == specifiers.DC_CURRENT_AMPLIFICATION:
        tex = r"\beta_{\mathrm{DC}}"
    elif self.specifier == specifiers.MAXIMUM_OSCILLATION_FREQUENCY:
        tex = r"f_{\mathrm{max" + subscript + r"}}"
    elif self.specifier == specifiers.TRANSIT_TIME:
        tex = r"\tau_{\mathrm{f" + subscript + r"}}"
    elif self.specifier == specifiers.FREQUENCY:
        tex = r"f_{\mathrm{" + subscript + r"}}"
    elif self.specifier == specifiers.TRANSCONDUCTANCE:
        tex = r"g_{\mathrm{m," + subscript + r"}}"
    elif self.specifier == specifiers.TEMPERATURE:
        tex = r"T_{\mathrm{" + subscript + r"}}"
    elif self.specifier == specifiers.NET_DOPING:
        tex = r"N_{\mathrm{net}}"
    elif self.specifier == specifiers.ACCEPTORS:
        tex = r"N_{\mathrm{A}}"
    elif self.specifier == specifiers.DONNORS:
        tex = r"N_{\mathrm{D}}"
    elif self.specifier == specifiers.TIME:
        if subscript:
            tex = r"t_{\mathrm{" + subscript + r"}}"
        else:
            tex = r"t"
    else:  # general case
        if subscript:
            tex = (
                str(self.specifier)
                + r"_{\mathrm{"
                + str("".join(self.nodes))
                + r","
                + subscript
                + r"}}"
            )
        else:
            tex = str(self.specifier) + r"_{\mathrm{" + str("".join(self.nodes)) + r"}}"

    # add superscript
    if superscript:
        tex = tex + r"^{" + superscript + r"}"

    # add sub_specifiers that wrap the main specifier
    if sub_specifiers.REAL.sub_specifiers <= self.sub_specifiers:
        tex = r"\Re\{" + tex + r"\}"
    elif sub_specifiers.IMAG.sub_specifiers <= self.sub_specifiers:
        tex = r"\Im\{" + tex + r"\}"
    elif sub_specifiers.MAG.sub_specifiers <= self.sub_specifiers:
        tex = r"|" + tex + r"|"
    elif sub_specifiers.PHASE.sub_specifiers <= self.sub_specifiers:
        tex = r"\angle\{" + tex + r"\}"

    # add specifiers that are appended to the string
    for specifier in self.sub_specifiers:
        if "|" + specifier in [
            sub_specifiers.PERIMETER,
            sub_specifiers.AREA,
            sub_specifiers.CORNER,
            sub_specifiers.LENGTH,
            sub_specifiers.WIDTH,
        ]:
            tex = tex + "|" + specifier

    return tex


SpecifierStr.to_tex = to_tex


def set_col_name(specifier, *nodes, sub_specifiers=None):
    """Returns the valid DMT name for this column.

    Parameters
    ----------
    specifier : str
        Ideally a specifier from :mod:`~DMT.core.specifiers`
    *nodes : str, optional
        Nodes for the column. Like 'B' for base or 'B', 'E' for base-emitter
    sub_specifiers : str, {'', :py:const:`DMT.core.sub_specifiers.PERIMETER`, :py:const:`DMT.core.sub_specifiers.AREA`}

    Returns
    -------
    col : str
        If nodes are given: specifier + '_' + ''.join(nodes)
        Without nodes: specifier
    """
    warnings.warn("Use SpecifierStr directly!", category=DeprecationWarning)

    return SpecifierStr(specifier, *nodes, sub_specifiers=sub_specifiers)


def get_nodes(string, nodes, fallback=None):
    """Return the nodes present in string in the correct order.

    Parameters
    ----------
    string          :  string
        String to be analyzed, e.g. 'V_BE'

    nodes           :  string
        Comma seperated list of possible nodes to be extracted.

    fallback        :  {string:string}
        Other nodes that may be present in string. If a key of this dict is found, the node name is mapped to its value.
        If value == '', the node is ignored.

    Returns
    -------
    relevant_nodes  :  list of string
        list of strings with the nodes in string.
    """
    # cast nodes from string to list of strings and order by string length.
    if isinstance(nodes, str):
        nodes = nodes.split(",")

    # go through string and find the nodes in the string and their position
    # code also accounts for nodes that occur multiple times, e.g. 'V' in 'V_BxB' where 'B' and 'Bx' are nodes
    nodes_ = copy.deepcopy(nodes)
    if fallback is not None:
        nodes_ = list(fallback.keys()) + nodes_
        nodes_ = set(nodes_)
        nodes_ = list(nodes_)

    nodes_.sort(key=len, reverse=True)  # order nodes by number of characters in node name
    pos_underscore = string.find("_")
    if pos_underscore == -1:  # if no underscore found, assume first letter is the specifier
        pos_underscore = 0
    string_ = string[pos_underscore + 1 :]  # throw away type specifier, i.e. 'C_'
    pos_sub_specifiers = string_.find("|")
    if pos_sub_specifiers != -1:  # sub_specifiers are optional
        string_ = string_[:pos_sub_specifiers]  # throw away sub_specifiers

    relevant_nodes_dict = {}
    for node in nodes_:
        while True:
            pos = string_.find(node)
            if pos != -1:
                relevant_nodes_dict[pos] = node
                string_ = string_.replace(node, "ä" * len(node), 1)
            else:
                break

    if string_.replace("ä", ""):
        # if string_ not empty, then the index is not only build of nodes -> do nothing!
        return ""

    relevant_nodes = list(relevant_nodes_dict.keys())  # for correct length
    for key, node in relevant_nodes_dict.items():
        relevant_nodes[key] = node

    # replace nodes from fallback dictionary with the desired node names
    if fallback is not None:
        for i, node in enumerate(relevant_nodes):
            if node in fallback.keys():
                if fallback[node] == "":
                    relevant_nodes[i] = node
                else:
                    relevant_nodes[i] = fallback[node]

    relevant_nodes = [node for node in relevant_nodes if node != ""]  # delete empty nodes

    return relevant_nodes


def get_sub_specifiers(string):
    """Returns the present sub_specifiers in string

    Parameters
    ----------
    string     : str
        The specifier of type str that is to be analyzed.

    Returns
    -------
    sub_specifiers : {str}
        A list of the sub_specifiers in specifier in the correct order.
    """
    relevant_subspecifiers = set()
    for sub_specifier in SUB_SPECIFIERS_STR:
        if sub_specifier in string:
            relevant_subspecifiers.add(sub_specifier)
            string = string.replace(sub_specifier, "")  # delete the ones which are already found

    return relevant_subspecifiers


def get_specifier_from_string(string, nodes=None):
    """Try to create a SpecifierStr object from a given string.

    If method fails, string is returned as is.
    if a string is given, which is already a specifier string, it is also returned as is.

    Parameters
    ----------
    string : str
        The string that shall be converted to SpecifierStr
    nodes  : iterable(str), optional
        An iterable that contains the nodes that shall be looked for in string.
    """
    if isinstance(string, SpecifierStr):
        return string

    # find the correct specifier
    specifier_in_string = ""
    # specifiers_sorted = sorted(SPECIFIERS, key=lambda speci: len(speci), reverse=True) SPECIFIERS are always sorted now...
    for speci in specifiers:
        if string.startswith(speci):
            specifier_in_string = speci
            break

    # did we find the specifier?
    if str(specifier_in_string) == string:  # yes, and it is the whole string, nice
        return specifier_in_string
    elif not specifier_in_string:  # no :(
        return string

    # yes but there is more...

    # if nodes are given we can check them next
    nodes_in_string = []
    if nodes is not None:
        nodes_in_string = get_nodes(string, nodes)

    # sub specifiers?
    sub_specifiers_in_string = get_sub_specifiers(string)

    specifier = SpecifierStr(
        specifier_in_string, *nodes_in_string, sub_specifiers=sub_specifiers_in_string
    )

    # test if everything is inside the new specifier
    rest = string.replace(str(specifier_in_string), "")
    if nodes_in_string:
        rest = rest.replace("_" + "".join(nodes_in_string), "")
    if sub_specifiers_in_string:
        for sub_spec in sub_specifiers_in_string:
            rest = rest.replace("|" + sub_spec, "")

    if rest:
        return string
    else:
        return specifier


# a dict that holds natural scaling for a few specifiers. Convenient for Plotting.
natural_scales = {
    specifiers.VOLTAGE: 1,
    specifiers.MAXIMUM_STABLE_GAIN: 1,
    specifiers.CURRENT: 1e3,  # mA
    specifiers.CURRENT_DENSITY: 1e3 / (1e6 * 1e6),  # mA/um^2
    specifiers.MAXIMUM_OSCILLATION_FREQUENCY: 1e-9,
    specifiers.TRANSCONDUCTANCE: 1,
    specifiers.TRANSIT_FREQUENCY: 1e-9,
    specifiers.CAPACITANCE: 1e15,
    specifiers.CHARGE: 1e15,
    specifiers.SS_PARA_Y: 1e3,
    specifiers.DC_CURRENT_AMPLIFICATION: 1,
    specifiers.FREQUENCY: 1e-9,  # GHz
    specifiers.TEMPERATURE: 1,
    specifiers.X: 1e9,
    specifiers.POWER: 1e3,  # mW
    specifiers.RESISTANCE: 1,  # Ohm
    specifiers.ENERGY: 1,  # eV
    specifiers.UNILATERAL_GAIN: 1,
    specifiers.NET_DOPING: 1e-6,  # 1/cm^3
}



# DMT_core
# Copyright (C) from 2022  SemiMod
# Copyright (C) until 2021  Markus Müller, Mario Krattenmacher and Pascal Kuthe
# <https://gitlab.com/dmt-development/dmt-core>
#
# This file is part of DMT_core.
#
# DMT_core is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DMT_core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>

from pint import UnitRegistry
from pathlib import Path
from pkg_resources import get_distribution

version_pkg = get_distribution("DMT-core").version.split(".")
try:
    from semver.version import Version as VersionInfo
except ImportError:
    from semver import VersionInfo

if "rc" in version_pkg[2]:
    _patch = version_pkg[2].split("rc")[0]
    _rc = version_pkg[2][len(_patch) + 2 :]
    __version__ = VersionInfo(
        major=version_pkg[0],
        minor=version_pkg[1],
        patch=_patch,
        prerelease="rc." + _rc,
    )
else:
    __version__ = VersionInfo(major=version_pkg[0], minor=version_pkg[1], patch=version_pkg[2])

# to get the next version:
# __version__.next_version(x) - with x = "major", "minor", "patch", "prerelease"
# or
# __version__.bump_x() - with x = "major", "minor", "patch", "prerelease"

name = "core"


from . import constants

# Helper
from .singleton import Singleton
from .hasher import create_md5_hash

# one unit registry for all of DMT

unit_registry = UnitRegistry()
from DMT.config import DATA_CONFIG


path_core = Path(__file__).resolve().parent
unit_registry.load_definitions(str(path_core / "dmt_units.txt"))

# helper for model equations and mcard memoization
from .utils import print_progress_bar
from .utils import enumerate_reversed

# naming conventions
from .naming import specifiers
from .naming import sub_specifiers
from .naming import specifiers_ss_para
from .naming import get_specifier_from_string
from .naming import get_nodes
from .naming import set_col_name
from .naming import SpecifierStr
from .naming import get_sub_specifiers
from .naming import natural_scales

# compact modelling stuff
from .mc_parameter import McParameter, McParameterCollection, McParameterComposition
from .va_file import VAFile
from .mcard import MCard
from .technology import Technology
from .circuit import Circuit, CircuitElement
from .mc_skywater import McSkywater

# plotting
from .plot import Plot
from .plot import save_or_show
from .plot import COMPARISON_3
from .plot_smith import SmithPlot
from .plot_2yaxis import Plot2YAxis

# Data management and processing
from .data_processor import is_iterable, flatten, strictly_increasing, DataProcessor

from .data_frame import DataFrame
from .sweep_def import SweepDef
from .sweep import Sweep, get_sweepdef
from .database_manager import DatabaseManager
from .data_reader import (
    read_data,
    save_elpa,
    read_ADS_bin,
    read_DEVICE_bin,
    read_elpa,
    read_mdm,
    read_hdf,
    read_feather,
)


# Simulation management
from .sim_con import SimCon

# DutView tree
from .dut_view import DutView
from .dut_type import DutType
from .dut_lib import DutLib

from .dut_meas import DutMeas
from .dut_dummy import DutDummy

from .dut_circuit import DutCircuit

from .dut_tcad import DutTcad

# docu
from .docu_dut_lib import DocuDutLib

# determine which modules are present
core_exists = True  # always, without DMT is not possible
try:
    import DMT.extraction

    extraction_exists = True
except ImportError:
    extraction_exists = False


if core_exists and not extraction_exists:
    mode = "free version"
elif core_exists and extraction_exists:
    mode = "all"

print("-----------------------------------------------------------------------")
print("DMT Copyright (C) 2022 SemiMod")
print("This program comes with ABSOLUTELY NO WARRANTY.")
print("DMT_core is free software, and you are welcome to redistribute it.")
if extraction_exists:
    print("DMT_other is free for non-commercial use, see LICENSE.md. ")
print("-----------------------------------------------------------------------")



import logging
from pathlib import Path
from DMT.core import DutMeas, DutType, DocuDutLib, DutLib, specifiers, sub_specifiers

folder_path = Path(__file__).resolve().parent
test_path = folder_path.parent
# -->Start main function
# --->Setup for log
logging.basicConfig(
    level=logging.DEBUG,
    format="%(levelname)s - %(message)s",
    filename=folder_path.parent.parent / "logs" / "test_dut_lib.log",
    filemode="w",
)


def create_lib():
    """Create a dut lib."""
    # -->Define subroutine at first
    # --->Subroutine: filter_dut
    def filter_dut(dut_name):
        if "dummies" in dut_name:
            return None
        elif "TLM" in dut_name:
            return None
        elif dut_name == "":
            return None
        else:
            dut_transistor = DutMeas(
                database_dir=test_path / "tmp",
                dut_type=DutType.npn,
                force=True,
                wafer=96,
                die="y",
                width=float(0.25e-6),
                length=float(0.25e-6),
                contact_config="CBEBC",
                name="dut_meas_npn",
                reference_node="E",
            )

            return dut_transistor

    # --->Arrange DMT class to dmt
    lib = DutLib(
        save_dir=test_path / "tmp",
        force=True,
        AC_filter_names=[("freq_vbc", "ac")],
        DC_filter_names=[("fgummel", "dc")],
    )
    # --->Add source measurement information in dmt and to duts
    lib.import_directory(
        import_dir=folder_path / "test_data",
        dut_filter=filter_dut,
        dut_level=1,
        force=True,
    )
    lib.dut_ref = lib.duts[0]

    # --->Add the open and short dummies to the lib
    dut_short = DutMeas(
        database_dir=test_path / "tmp",
        name="dut_short_npn",
        dut_type=DutType.deem_short_bjt,
        reference_node="E",
        force=True,
        wafer=96,
        die="y",
        width=float(0.25e-6),
        length=float(0.25e-6),
    )
    dut_short.add_data(folder_path / "test_data" / "dummy_short_freq.mdm", key="ac")
    dut_short.add_data(folder_path / "test_data" / "short_dc.mdm", key="dc")
    dut_open = DutMeas(
        database_dir=test_path / "tmp",
        name="dut_open_npn",
        dut_type=DutType.deem_open_bjt,
        reference_node="E",
        force=True,
        wafer=96,
        die="y",
        width=float(0.25e-6),
        length=float(0.25e-6),
    )
    dut_open.add_data(folder_path / "test_data" / "dummy_open_freq.mdm", key="ac")
    lib.add_duts([dut_short, dut_open])

    # --->Clean the names of all dataframes, e.g. VB=>V_B
    for dut_a in lib:
        dut_a.clean_data(
            fallback={
                "S_DEEMB(1,1)": None,
                "S_DEEMB(2,1)": None,
                "S_DEEMB(1,2)": None,
                "S_DEEMB(2,2)": None,
            }
        )

    # emulate forced specifiers
    col_vb = specifiers.VOLTAGE + "B"
    col_ve = specifiers.VOLTAGE + "E"
    col_vc = specifiers.VOLTAGE + "C"
    for key in lib.dut_ref.data.keys():
        lib.dut_ref.data[key][col_vb + sub_specifiers.FORCED] = lib.dut_ref.data[key][col_vb]
        lib.dut_ref.data[key][col_vc + sub_specifiers.FORCED] = lib.dut_ref.data[key][col_vc]
        lib.dut_ref.data[key][col_ve + sub_specifiers.FORCED] = lib.dut_ref.data[key][col_ve]

    return lib


def test_docu():

    lib_test = create_lib()
    docu = DocuDutLib(lib_test)

    docu.generate_docu(
        folder_path.parent / "tmp" / "docu_dut_lib",
        plot_specs=[{"type": "gummel_vbc", "key": "fgummel"}],
        show=False,  # not possible in CI/CD
        save_tikz_settings={
            "width": "3in",
            "height": "5in",
            "standalone": True,
            "svg": False,
            "build": False,  # not possible in CI/CD
            "mark_repeat": 20,
            "clean": False,  # Remove all files except *.pdf files in plots
        },
    )


if __name__ == "__main__":
    lib_test = test_docu()



""" test ngspice vs ADS using a real modelcard in a systematic way.
"""
import numpy as np
import copy
import logging
from pathlib import Path
from DMT.core import SimCon, Plot, DutType, Sweep, specifiers, McParameter, sub_specifiers
from DMT.ngspice import DutNgspice
from DMT.ADS import DutAds
from DMT.hl2 import McHicum, Hl2Model, VA_FILES


folder_path = Path(__file__).resolve().parent
test_path = folder_path.parent

logging.basicConfig(
    level=logging.DEBUG,
    format="%(levelname)s - %(message)s",
    filename=folder_path.parent.parent / "logs" / "test_ADS_systematic.log",
    filemode="w",
)


def compare_ngspice_ads_real_mc(mc):

    rth_de = McParameter(name="rth_de", value=0)
    try:
        mc.add(rth_de)
    except:
        pass

    # mc = McHicum(
    #     load_model_from_path=os.path.join('test', 'test_interface_ngspice', 'dietmar_output.lib'),
    # )
    # mc.va_file = VA_FILES['L2V2.4.0_internal']
    # mc.set_values({
    #     'alrth':0,
    #     'flsh':0,
    # })

    dut_ads = DutAds(
        None, DutType.npn, mc, nodes="C,B,E,S,T", reference_node="E"
    )  # ADS with VA file

    # create a sweep with ONE operating point
    sweepdef = [
        {
            "var_name": specifiers.FREQUENCY,
            "sweep_order": 4,
            "sweep_type": "CON",
            "value_def": [1e9],
        },
        # {'var_name':specifiers.VOLTAGE + 'B',  'sweep_order':3, 'sweep_type':'CON' , 'value_def':[0.85]},
        # {'var_name':specifiers.VOLTAGE + 'B',  'sweep_order':3, 'sweep_type':'CON' , 'value_def':[-2.0]},
        {
            "var_name": specifiers.VOLTAGE + "B",
            "sweep_order": 3,
            "sweep_type": "LIN",
            "value_def": [0.3, 1, 91],
        },
        # {'var_name':specifiers.VOLTAGE + 'B',  'sweep_order':3, 'sweep_type':'CON' , 'value_def':[0.9]},
        {
            "var_name": specifiers.VOLTAGE + "C",
            "sweep_order": 2,
            "sweep_type": "CON",
            "value_def": [2],
        },
        {
            "var_name": specifiers.VOLTAGE + "E",
            "sweep_order": 1,
            "sweep_type": "CON",
            "value_def": [0],
        },
        # {'var_name':specifiers.VOLTAGE + 'B',  'sweep_order':3, 'sweep_type':'CON' , 'value_def':[0.9]},
        # {'var_name':specifiers.VOLTAGE + 'C',  'sweep_order':2, 'sweep_type':'CON' , 'value_def':[2.]},
        # {'var_name':specifiers.VOLTAGE + 'E',  'sweep_order':1, 'sweep_type':'CON' , 'value_def':[0]},
    ]
    outputdef = ["I_C", "I_B"]
    othervar = {specifiers.TEMPERATURE: 398}
    sweep = Sweep("gummel", sweepdef=sweepdef, outputdef=outputdef, othervar=othervar)

    # first simulate ADS
    sim_con = SimCon(n_core=1, t_max=30)
    sim_con.append_simulation(dut=dut_ads, sweep=sweep)
    sim_con.run_and_read(force=True)

    # read initial conditions
    data_ads = dut_ads.get_data(sweep=sweep)
    ve = 0
    vb = data_ads["VBE"].to_numpy()[0]
    vc = data_ads["VCE"].to_numpy()[0]
    vbi = data_ads["VBiE"].to_numpy()[0]
    vbp = data_ads["VBpE"].to_numpy()[0]
    vci = data_ads["VCiE"].to_numpy()[0]
    vei = data_ads["VEiE"].to_numpy()[0]
    vt = data_ads["DTSH"].to_numpy()[0]
    delta = 1
    initial_conditions = {
        "icVE": ve * delta,
        "icVEi": vei * delta,
        "icVC": vc * delta,
        "icVCi": vci * delta,
        "icVB": vb * delta,
        "icVBi": vbi * delta,
        "icVBp": vbp * delta,
        "icVt": vt * delta,
    }

    # get rth stuff from ads
    rth_t = data_ads["rth_t"].to_numpy()
    pterm = data_ads["pterm"].to_numpy()
    dT = rth_t * pterm

    assert np.isclose(vt, dT[0], rtol=1e-3)

    # simulate ngspice using initial condition from ADS
    dut_ngspice = DutNgspice(
        None,
        DutType.npn,
        mc,
        nodes="C,B,E,S,T",
        reference_node="E",
        copy_va_files=False,  # ngspice without VA file! -> use internal
        simulator_command="ngspice",
        initial_conditions=initial_conditions,
        simulator_options={"GMIN": 1e-18, "ABSTOL": 1e-15},
    )
    sim_con.append_simulation(dut=dut_ngspice, sweep=sweep)
    sim_con.run_and_read(force=True)

    data_ngspice = dut_ngspice.get_data(sweep=sweep)

    # ngspice operating point
    vbiei = data_ngspice["VBIEI"].to_numpy()
    vbici = data_ngspice["VBICI"].to_numpy()
    vciei = data_ngspice["VCIEI"].to_numpy()
    temp = data_ngspice["TK"].to_numpy()
    # ngspice currents
    ibei_ngspice = data_ngspice["IBEI"].to_numpy()
    ibei_ads = data_ads["IBIEI"].to_numpy()
    it_ngspice = data_ngspice["IT"].to_numpy()
    it_ads = data_ads["IT_"].to_numpy()

    # ensure some stuff
    ensure = [
        specifiers.TRANSIT_FREQUENCY,
        specifiers.CAPACITANCE + "B" + "E",
        specifiers.CAPACITANCE + "B" + "C",
        specifiers.CAPACITANCE + "C" + "E",
        specifiers.SS_PARA_Y + "B" + "B" + sub_specifiers.IMAG,
        specifiers.SS_PARA_Y + "B" + "B" + sub_specifiers.REAL,
        specifiers.SS_PARA_Y + "C" + "B" + sub_specifiers.REAL,
        specifiers.SS_PARA_Y + "C" + "B" + sub_specifiers.IMAG,
        specifiers.SS_PARA_Y + "B" + "C" + sub_specifiers.REAL,
        specifiers.SS_PARA_Y + "B" + "C" + sub_specifiers.IMAG,
        specifiers.SS_PARA_Y + "C" + "C" + sub_specifiers.REAL,
        specifiers.SS_PARA_Y + "C" + "C" + sub_specifiers.IMAG,
    ]
    for speci in ensure:
        data_ngspice.ensure_specifier_column(speci, ports=["B", "C"])
        data_ads.ensure_specifier_column(speci, ports=["B", "C"])

    for col in data_ngspice.columns:
        col_data = data_ngspice[col].to_numpy()
        if np.all(np.imag(col_data) == 0):
            data_ngspice[col] = data_ngspice[col].astype(float)

    return data_ngspice, data_ads


if __name__ == "__main__":
    show = True  # If True, plot in maptplotlib, else generate Tex

    # load a modelcard
    mc = McHicum(load_model_from_path=folder_path / "hicum_test_modelcard.lib")

    mc = mc.get_clean_modelcard()
    mc.set_values(
        {"flcomp": 2.4, "flnqs": 1, "flsh": 1, "alrth": 0, "cscp0": 2e-15, "cth": 1, "rbi0": 0},
        force=True,
    )
    mc.va_file = VA_FILES["L2V2.4.0_internal"]
    # end load modelcard

    # define plots
    plt_gummel = Plot(
        r"$I(V_{\mathrm{BE}})$",
        x_specifier=specifiers.VOLTAGE + "B" + "E",
        y_specifier=specifiers.CURRENT + "C",
        y_log=True,
        style="comparison_4",
    )
    plt_it = Plot(
        r"IT(VBE)",
        x_specifier=specifiers.VOLTAGE + "B" + "E",
        y_specifier=specifiers.CURRENT + "C",
        y_log=True,
        style="xtraction_color",
    )
    plt_temp = Plot(
        r"$dT(VBE)$",
        x_specifier=specifiers.VOLTAGE + "B" + "E",
        y_specifier=specifiers.TEMPERATURE,
        style="xtraction_color",
    )
    plt_ft = Plot(
        r"FT(VBE)",
        x_specifier=specifiers.CURRENT + "C",
        y_specifier=specifiers.TRANSIT_FREQUENCY,
        x_log=True,
        y_scale=1e-9,
        style="xtraction_color",
    )
    plt_ft.x_limits = (1e-5, 1e-1)
    plt_ft.y_limits = (0, 700)
    plt_cbe = Plot(
        r"CBE(VBE)",
        x_specifier=specifiers.VOLTAGE + "B" + "E",
        y_specifier=specifiers.CAPACITANCE + "B" + "E",
        y_scale=1e15,
        style="xtraction_color",
    )
    plt_cbc = Plot(
        r"CBC(VBE)",
        x_specifier=specifiers.VOLTAGE + "B" + "E",
        y_specifier=specifiers.CAPACITANCE + "B" + "C",
        y_scale=1e15,
        style="xtraction_color",
    )
    plt_cce = Plot(
        r"CCE(VBE)",
        x_specifier=specifiers.VOLTAGE + "B" + "E",
        y_specifier=specifiers.CAPACITANCE + "C" + "E",
        y_scale=1e15,
        style="xtraction_color",
    )
    plt_vbici = Plot(
        r"VBICI(VBE)",
        x_specifier=specifiers.VOLTAGE + "B",
        y_specifier=specifiers.VOLTAGE + "Bi" + "Ci",
        style="xtraction_color",
    )
    plt_rbi = Plot(
        r"rbi(VBE)",
        x_specifier=specifiers.VOLTAGE + "B",
        y_label=r"$r_{\mathrm{rbi}}$",
        style="xtraction_color",
    )
    plt_rey11 = Plot(
        r"ReY11(VBE)",
        x_specifier=specifiers.VOLTAGE + "B",
        y_specifier=specifiers.SS_PARA_Y + "B" + "B" + sub_specifiers.REAL,
        style="xtraction_color",
    )
    plt_imy11 = Plot(
        r"ImY11(VBE)",
        x_specifier=specifiers.VOLTAGE + "B",
        y_specifier=specifiers.SS_PARA_Y + "B" + "B" + sub_specifiers.IMAG,
        style="xtraction_color",
    )
    plt_rey21 = Plot(
        r"ReY21(VBE)",
        x_specifier=specifiers.VOLTAGE + "B",
        y_specifier=specifiers.SS_PARA_Y + "C" + "B" + sub_specifiers.REAL,
        style="markers_lines",
        y_log=True,
    )
    plt_imy21 = Plot(
        r"ImY21(VBE)",
        x_specifier=specifiers.VOLTAGE + "B",
        y_specifier=specifiers.SS_PARA_Y + "C" + "B" + sub_specifiers.IMAG,
        y_log=True,
        style="xtraction_color",
    )
    plt_rey12 = Plot(
        r"ReY12(VBE)",
        x_specifier=specifiers.VOLTAGE + "B",
        y_specifier=specifiers.SS_PARA_Y + "B" + "C" + sub_specifiers.REAL,
        style="xtraction_color",
    )
    plt_imy12 = Plot(
        r"ImY12(VBE)",
        x_specifier=specifiers.VOLTAGE + "B",
        y_specifier=specifiers.SS_PARA_Y + "B" + "C" + sub_specifiers.IMAG,
        y_log=True,
        style="xtraction_color",
    )
    plt_rey22 = Plot(
        r"ReY22(VBE)",
        x_specifier=specifiers.VOLTAGE + "B",
        y_specifier=specifiers.SS_PARA_Y + "C" + "C" + sub_specifiers.REAL,
        style="xtraction_color",
        y_log=True,
    )
    plt_imy22 = Plot(
        r"ImY22(VBE)",
        x_specifier=specifiers.VOLTAGE + "B",
        y_specifier=specifiers.SS_PARA_Y + "C" + "C" + sub_specifiers.IMAG,
        y_log=True,
        style="xtraction_color",
    )
    plt_vbiei = Plot(
        r"VBIEI(VBE)",
        x_specifier=specifiers.VOLTAGE + "B",
        y_specifier=specifiers.VOLTAGE + "Bi" + "Ei",
        style="xtraction_color",
    )
    # end define plots

    # store all reactive elements value
    caps = [
        "cjcx0",
        "cjci0",
        "cjep0",
        "cjei0",
        "cbcpar",
        # 'cbepar',
        # 'cjs0',
        # 'fcrbi',
        # 'csu',
        # 'cscp0',
        "all",
    ]

    for cap in caps:
        mc_modified = copy.deepcopy(mc)
        try:
            cap_val = mc[cap].value
            for caps_ in caps[:-2]:
                mc_modified.set_values({caps_: 0})
            mc_modified.set_values({cap: cap_val})
            label = " only " + cap
        except:
            dummy = 1
            pass
        if cap == "all":
            label = "all on"

        data_ngspice, data_ads = compare_ngspice_ads_real_mc(mc_modified)

        # fill plots
        # gummel
        plt_gummel.add_data_set(data_ngspice["V_B"], data_ngspice["I_C"], label="ngspice " + label)
        plt_gummel.add_data_set(data_ads["V_B"], data_ads["I_C"], label="ADS " + label)
        plt_gummel.add_data_set(
            data_ngspice["V_B"],
            data_ngspice["I_B"],  # IB from ngspice short is wrong
            # data_ngspice['VBBP']/mc['rbx'].value,
            # label='ngspice ' + label
        )
        # plt_gummel.add_data_set(
        #     data_ngspice['V_B'],
        #     data_ngspice['IBEI'], #IB from ngspice short is wrong
        #     #data_ngspice['VBBP']/mc['rbx'].value,
        #     label='ngspice $IBEI$'
        # )
        plt_gummel.add_data_set(
            data_ads["V_B"],
            data_ads["I_B"],
            # label='ADS ' + label
        )
        # it
        plt_it.add_data_set(data_ngspice["V_B"], data_ngspice["IT"], label="ngspice " + label)
        plt_it.add_data_set(data_ads["V_B"], data_ads["IT_"], label="ADS " + label)
        # rbi
        plt_rbi.add_data_set(data_ngspice["V_B"], data_ngspice["RBI"], label="ngspice " + label)
        plt_rbi.add_data_set(data_ads["V_B"], data_ads["rbi"], label="ADS " + label)
        # temp
        plt_temp.add_data_set(data_ngspice["V_B"], data_ngspice["V_T"], label="ngspice " + label)
        plt_temp.add_data_set(data_ads["V_B"], data_ads["DTSH"], label="ADS " + label)
        # ft
        try:
            plt_ft.add_data_set(
                data_ngspice["I_C"],
                data_ngspice[specifiers.TRANSIT_FREQUENCY],
                label="ngspice " + label,
            )
            plt_ft.add_data_set(
                data_ads["I_C"], data_ads[specifiers.TRANSIT_FREQUENCY], label="ADS " + label
            )
        except KeyError:
            pass
        # cbe
        try:
            plt_cbe.add_data_set(
                data_ngspice["V_B"],
                data_ngspice[specifiers.CAPACITANCE + "B" + "E"],
                label="ngspice " + label,
            )
            plt_cbe.add_data_set(
                data_ads["V_B"], data_ads[specifiers.CAPACITANCE + "B" + "E"], label="ADS " + label
            )
        except KeyError:
            pass
        # cbc
        try:
            plt_cbc.add_data_set(
                data_ngspice["V_B"],
                data_ngspice[specifiers.CAPACITANCE + "B" + "C"],
                label="ngspice " + label,
            )
            plt_cbc.add_data_set(
                data_ads["V_B"], data_ads[specifiers.CAPACITANCE + "B" + "C"], label="ADS " + label
            )
        except KeyError:
            pass
        # cce
        try:
            plt_cce.add_data_set(
                data_ngspice["V_B"],
                data_ngspice[specifiers.CAPACITANCE + "C" + "E"],
                label="ngspice " + label,
            )
            plt_cce.add_data_set(
                data_ads["V_B"], data_ads[specifiers.CAPACITANCE + "C" + "E"], label="ADS " + label
            )
        except KeyError:
            pass
        # reY11
        try:
            plt_rey11.add_data_set(
                data_ngspice["V_B"],
                data_ngspice[specifiers.SS_PARA_Y + "B" + "B" + sub_specifiers.REAL],
                label="ngspice " + label,
            )
            plt_rey11.add_data_set(
                data_ads["V_B"],
                data_ads[specifiers.SS_PARA_Y + "B" + "B" + sub_specifiers.REAL],
                label="ADS " + label,
            )
        except KeyError:
            pass
        # imY11
        try:
            plt_imy11.add_data_set(
                data_ngspice["V_B"],
                data_ngspice[specifiers.SS_PARA_Y + "B" + "B" + sub_specifiers.IMAG],
                label="ngspice " + label,
            )
            plt_imy11.add_data_set(
                data_ads["V_B"],
                data_ads[specifiers.SS_PARA_Y + "B" + "B" + sub_specifiers.IMAG],
                label="ADS " + label,
            )
        except KeyError:
            pass
        # reY21
        try:
            plt_rey21.add_data_set(
                data_ngspice["V_B"],
                data_ngspice[specifiers.SS_PARA_Y + "C" + "B" + sub_specifiers.REAL],
                label="ngspice " + label,
            )
            plt_rey21.add_data_set(
                data_ads["V_B"],
                data_ads[specifiers.SS_PARA_Y + "C" + "B" + sub_specifiers.REAL],
                label="ADS " + label,
            )
        except KeyError:
            pass
        # imY21
        try:
            plt_imy21.add_data_set(
                data_ngspice["V_B"],
                data_ngspice[specifiers.SS_PARA_Y + "C" + "B" + sub_specifiers.IMAG],
                label="ngspice " + label,
            )
            plt_imy21.add_data_set(
                data_ads["V_B"],
                data_ads[specifiers.SS_PARA_Y + "C" + "B" + sub_specifiers.IMAG],
                label="ADS " + label,
            )
        except KeyError:
            pass
        # reY12
        try:
            plt_rey12.add_data_set(
                data_ngspice["V_B"],
                data_ngspice[specifiers.SS_PARA_Y + "B" + "C" + sub_specifiers.REAL],
                label="ngspice " + label,
            )
            plt_rey12.add_data_set(
                data_ads["V_B"],
                data_ads[specifiers.SS_PARA_Y + "B" + "C" + sub_specifiers.REAL],
                label="ADS " + label,
            )
        except KeyError:
            pass
        # imY12
        try:
            plt_imy12.add_data_set(
                data_ngspice["V_B"],
                data_ngspice[specifiers.SS_PARA_Y + "B" + "C" + sub_specifiers.IMAG],
                label="ngspice " + label,
            )
            plt_imy12.add_data_set(
                data_ads["V_B"],
                data_ads[specifiers.SS_PARA_Y + "B" + "C" + sub_specifiers.IMAG],
                label="ADS " + label,
            )
        except KeyError:
            pass
        # reY22
        try:
            plt_rey22.add_data_set(
                data_ngspice["V_B"],
                data_ngspice[specifiers.SS_PARA_Y + "C" + "C" + sub_specifiers.REAL],
                label="ngspice " + label,
            )
            plt_rey22.add_data_set(
                data_ads["V_B"],
                data_ads[specifiers.SS_PARA_Y + "C" + "C" + sub_specifiers.REAL],
                label="ADS " + label,
            )
        except KeyError:
            pass
        # imY22
        try:
            plt_imy22.add_data_set(
                data_ngspice["V_B"],
                data_ngspice[specifiers.SS_PARA_Y + "C" + "C" + sub_specifiers.IMAG],
                label="ngspice " + label,
            )
            plt_imy22.add_data_set(
                data_ads["V_B"],
                data_ads[specifiers.SS_PARA_Y + "C" + "C" + sub_specifiers.IMAG],
                label="ADS " + label,
            )
        except KeyError:
            pass
        # vbici
        plt_vbici.add_data_set(data_ngspice["V_B"], data_ngspice["VBICI"], label="ngspice " + label)
        plt_vbici.add_data_set(
            data_ads["V_B"], data_ads["_V_BiE"] - data_ads["_V_CiE"], label="ADS " + label
        )
        # vbiei
        plt_vbiei.add_data_set(data_ngspice["V_B"], data_ngspice["VBIEI"], label="ngspice " + label)
        plt_vbiei.add_data_set(
            data_ads["V_B"], data_ads["_V_BiE"] - data_ads["_V_EiE"], label="ADS " + label
        )

    plts = [
        plt_vbiei,
        plt_rbi,
        plt_vbici,
        plt_rey11,
        plt_imy11,
        plt_rey12,
        plt_imy12,
        plt_rey21,
        plt_imy21,
        plt_rey22,
        plt_imy22,
        plt_ft,
        plt_cbe,
        plt_cbc,
        plt_cce,
        plt_temp,
        plt_it,
        plt_gummel,
    ]
    if show:
        for plt in plts[:-2]:
            plt.plot_py(show=False)

        plts[-1].plot_py(show=True)

    else:
        ###PRINT TO DOCUMENT FROM HERE#####################
        from pylatex import (
            Section,
            Subsection,
            Tabular,
            Math,
            Plot,
            Figure,
            Matrix,
            Alignat,
            Package,
            Itemize,
            NoEscape,
            Document,
        )
        from pylatex.base_classes import Arguments
        from pylatexenc.latex2text import LatexNodes2Text

        # from DMT.gui.utils import SubFile, CommandRefRange, CommandRef, CommandInput, CommandLabel from DMT.gui.utils import SubFile, CommandRefRange, CommandRef, CommandInput, CommandLabel from pylatexenc.latex2text import LatexNodes2Text
        from DMT.external import SubFile, CommandRef, CommandInput, CommandLabel

        width = "4.5in"

        doc = Document()
        fig_dir = folder_path / "documentation", "figs"

        with doc.create(Section("Comparison ADS/NGspice")):
            doc.append(
                NoEscape(
                    r'In this document the ngspice HICUM/L2.4.3 model in ngspice is compared against ADS simulations. The modelcard is taken from a real process and is realistic. This document is auto-generated using Pylatex. The shown ft, CBE, CCE and CBC results show quantities that are calculated from simulated Y-parameters. In these simulations all reactive model elements but one are turned off (except those simulations labeled as "all"). E.g. "only cjei0" means that only the Cjei capacitance is active.'
                )
            )
            with doc.create(Subsection("Plots")):
                for plt in plts:
                    plt_name = plt.save_tikz(
                        fig_dir, width=width, mark_repeat=2, legend_location="upper right outer"
                    )
                    with doc.create(Figure(position="h!")) as _plot:
                        _plot.append(
                            CommandInput(arguments=Arguments('"' + str(fig_dir / plt_name) + '"'))
                        )
                        _plot.add_caption(NoEscape(plt.name))
                        _plot.append(
                            CommandLabel(
                                arguments=Arguments(LatexNodes2Text().latex_to_text(plt.name))
                            )
                        )

        doc.generate_tex(folder_path / "documentation" / "comparison_ngspice_ads")



""" test ngspice vs ADS using a real modelcard.
"""
import logging
from pathlib import Path
from DMT.core import SimCon, Plot, DutType, Sweep, specifiers, McParameter
from DMT.ngspice import DutNgspice
from DMT.ADS import DutAds
from DMT.hl2 import McHicum, Hl2Model, VA_FILES

import numpy as np

folder_path = Path(__file__).resolve().parent
test_path = folder_path.parent

logging.basicConfig(
    level=logging.DEBUG,
    format="%(levelname)s - %(message)s",
    filename=folder_path.parent.parent / "logs" / "test_ADS_real.log",
    filemode="w",
)


def compare_ngspice_ads_real_mc():
    mc = McHicum.load(folder_path / "real_mcard.mcard")

    ###### Disable branches
    ## General settings
    mc = mc.get_clean_modelcard()
    mc.set_values(
        {
            "flcomp": 2.4,
            "flnqs": 0,
            "flsh": 1,
            "alrth": 0,
            # 'rth'   : 3700,
            # 'kavl'  : 0,
            # 'qavl': 10,
            # 'ibcis':0,
            # 'ibcxs':0,
            # 'ibeis':1.3e-18,
            # 'ireis':0,
            # 'ibets':0,
            # 'ibeps':0,
            # 'ireps':0,
            # 'itss'  :0,
            # 'rbx'   :10,
            # 'tr'    :0,
            # 'tbhrec':0,
            # 'rcx'   :1,
        },
        force=True,
    )
    mc.va_file = VA_FILES["L2V2.4.0_internal"]

    rth_de = McParameter(name="rth_de", value=0)
    mc.add(rth_de)

    # mc = McHicum(
    #     load_model_from_path=os.path.join('test', 'test_interface_ngspice', 'dietmar_output.lib'),
    # )
    # mc.va_file = VA_FILES['L2V2.4.0_internal']
    # mc.set_values({
    #     'alrth':0,
    #     'rth':1200,
    # })

    dut_ads = DutAds(
        None, DutType.npn, mc, nodes="C,B,E,S,T", reference_node="E"
    )  # ADS with VA file

    # create a sweep with ONE operating point
    sweepdef = [
        {
            "var_name": specifiers.FREQUENCY,
            "sweep_order": 4,
            "sweep_type": "CON",
            "value_def": [10e9],
        },
        # {'var_name':specifiers.VOLTAGE + 'B',  'sweep_order':3, 'sweep_type':'CON' , 'value_def':[0.85]},
        # {'var_name':specifiers.VOLTAGE + 'B',  'sweep_order':3, 'sweep_type':'CON' , 'value_def':[-2.0]},
        # {'var_name':specifiers.VOLTAGE + 'B',  'sweep_order':3, 'sweep_type':'LIN' , 'value_def':[0.3, 1, 91]},
        {
            "var_name": specifiers.VOLTAGE + "B",
            "sweep_order": 3,
            "sweep_type": "CON",
            "value_def": [0.9],
        },
        {
            "var_name": specifiers.VOLTAGE + "C",
            "sweep_order": 2,
            "sweep_type": "CON",
            "value_def": [2],
        },
        {
            "var_name": specifiers.VOLTAGE + "E",
            "sweep_order": 1,
            "sweep_type": "CON",
            "value_def": [0],
        },
        # {'var_name':specifiers.VOLTAGE + 'B',  'sweep_order':3, 'sweep_type':'CON' , 'value_def':[0.9]},
        # {'var_name':specifiers.VOLTAGE + 'C',  'sweep_order':2, 'sweep_type':'CON' , 'value_def':[2.]},
        # {'var_name':specifiers.VOLTAGE + 'E',  'sweep_order':1, 'sweep_type':'CON' , 'value_def':[0]},
    ]
    outputdef = ["I_C", "I_B"]
    othervar = {"TEMP": 300}
    sweep = Sweep("gummel", sweepdef=sweepdef, outputdef=outputdef, othervar=othervar)

    # create a sweep with ONE operating point
    sweepdef = [
        {
            "var_name": specifiers.FREQUENCY,
            "sweep_order": 4,
            "sweep_type": "CON",
            "value_def": [10e9],
        },
        # {'var_name':specifiers.VOLTAGE + 'B',  'sweep_order':3, 'sweep_type':'CON' , 'value_def':[0.85]},
        # {'var_name':specifiers.VOLTAGE + 'B',  'sweep_order':3, 'sweep_type':'CON' , 'value_def':[-2.0]},
        {
            "var_name": specifiers.VOLTAGE + "B",
            "sweep_order": 3,
            "sweep_type": "LIN",
            "value_def": [0.3, 1, 91],
        },
        # {'var_name':specifiers.VOLTAGE + 'B',  'sweep_order':3, 'sweep_type':'CON' , 'value_def':[0.9]},
        {
            "var_name": specifiers.VOLTAGE + "C",
            "sweep_order": 2,
            "sweep_type": "CON",
            "value_def": [2],
        },
        {
            "var_name": specifiers.VOLTAGE + "E",
            "sweep_order": 1,
            "sweep_type": "CON",
            "value_def": [0],
        },
        # {'var_name':specifiers.VOLTAGE + 'B',  'sweep_order':3, 'sweep_type':'CON' , 'value_def':[0.9]},
        # {'var_name':specifiers.VOLTAGE + 'C',  'sweep_order':2, 'sweep_type':'CON' , 'value_def':[2.]},
        # {'var_name':specifiers.VOLTAGE + 'E',  'sweep_order':1, 'sweep_type':'CON' , 'value_def':[0]},
    ]
    outputdef = ["I_C", "I_B"]
    othervar = {"TEMP": 300}
    sweep = Sweep("gummel", sweepdef=sweepdef, outputdef=outputdef, othervar=othervar)

    # first simulate ADS
    sim_con = SimCon(n_core=1, t_max=30)
    sim_con.append_simulation(dut=dut_ads, sweep=sweep)
    sim_con.run_and_read(force=False)

    # read initial conditions
    data_ads = dut_ads.get_data(sweep=sweep)
    ve = 0
    vb = data_ads["VBE"].to_numpy()[0]
    vc = data_ads["VCE"].to_numpy()[0]
    vbi = data_ads["VBiE"].to_numpy()[0]
    vbp = data_ads["VBpE"].to_numpy()[0]
    vci = data_ads["VCiE"].to_numpy()[0]
    vei = data_ads["VEiE"].to_numpy()[0]
    vt = data_ads["DTSH"].to_numpy()[0]
    delta = 1
    initial_conditions = {
        "icVE": ve * delta,
        "icVEi": vei * delta,
        "icVC": vc * delta,
        "icVCi": vci * delta,
        "icVB": vb * delta,
        "icVBi": vbi * delta,
        "icVBp": vbp * delta,
        "icVt": vt * delta,
    }

    # get rth stuff from ads
    rth_t = data_ads["rth_t"].to_numpy()
    pterm = data_ads["pterm"].to_numpy()
    dT = rth_t * pterm

    assert np.isclose(vt, dT[0], rtol=1e-3)

    # simulate ngspice using initial condition from ADS
    dut_ngspice = DutNgspice(
        None,
        DutType.npn,
        mc,
        nodes="C,B,E,S,T",
        reference_node="E",
        copy_va_files=False,  # ngspice without VA file! -> use internal
        simulator_command="ngspice",
        initial_conditions=initial_conditions,
        simulator_options={"GMIN": 1e-18, "ABSTOL": 1e-15},
    )
    sim_con.append_simulation(dut=dut_ngspice, sweep=sweep)
    sim_con.run_and_read(force=True)

    data_ngspice = dut_ngspice.get_data(sweep=sweep)
    try:
        data_ngspice.ensure_specifier_column(
            specifiers.TRANSIT_FREQUENCY, ports=dut_ngspice.ac_ports
        )
        data_ads.ensure_specifier_column(specifiers.TRANSIT_FREQUENCY, ports=dut_ads.ac_ports)
    except StopIteration:
        pass

    model = Hl2Model(model_name="hl2", version="2.4.3")
    kwargs = mc.to_kwargs()

    # ngspice operating point
    vbiei = data_ngspice["VBIEI"].to_numpy()
    vbici = data_ngspice["VBICI"].to_numpy()
    vciei = data_ngspice["VCIEI"].to_numpy()
    temp = data_ngspice["TK"].to_numpy()
    # ngspice currents
    ibei_ngspice = data_ngspice["IBEI"].to_numpy()
    ibei_ads = data_ads["IBIEI"].to_numpy()
    it_ngspice = data_ngspice["IT"].to_numpy()
    it_ads = data_ads["IT_"].to_numpy()
    # some assertions since I am at this point clueeless ... ##########################################
    # assert np.allclose(
    #     model.model_ibei_temp(vbe=vbiei, t_dev=temp,**kwargs),
    #     data_ngspice['IBEI'].to_numpy(),
    #     rtol=1e-2
    # )
    try:
        assert np.allclose(
            model.model_ibci_temp(vbc=vbici, t_dev=temp, **kwargs),
            data_ngspice["IBCI"].to_numpy(),
            rtol=1e-3,
        )  # not exactly zero due to gmin
    except AssertionError:
        pass  # gmin makes this assertion fail for small current
    assert np.allclose(
        model.model_ick_temp(vce=vciei, t_dev=temp, **kwargs),
        data_ngspice["ICK"].to_numpy(),
        rtol=1e-3,
    )  # not exactly zero due to gmin
    assert np.allclose(
        model.model_cjci_temp(vbc=vbici, t_dev=temp, **kwargs),
        data_ngspice["CMUI"].to_numpy(),
        rtol=1e-3,
    )
    # assert np.allclose(
    #     model.model_cjei_temp(vbe=vbiei, t_dev=temp, **kwargs),
    #     data_ngspice['CPII'].to_numpy(),
    #     rtol=1e-3
    # ) # does not work because of CdE

    # using currents

    # assert np.allclose(
    #     ibei_ngspice,
    #     ibei_ads,
    #     rtol=1e-3
    # )
    # assert np.allclose(
    #     it_ngspice,
    #     it_ads,
    #     rtol=1e-3
    # )

    return data_ngspice, data_ads


if __name__ == "__main__":
    data_ngspice, data_ads = compare_ngspice_ads_real_mc()

    # define plots
    plt_gummel = Plot(
        r"$I(V_{\mathrm{BE}})$",
        x_specifier=specifiers.VOLTAGE + "B" + "E",
        y_specifier=specifiers.CURRENT + "C",
        y_log=True,
        style="markers_lines",
        legend_location="upper left",
    )
    plt_ibei = Plot(
        r"$IBEI(V_{\mathrm{BE}})$",
        x_specifier=specifiers.VOLTAGE + "B" + "E",
        y_specifier=specifiers.CURRENT + "B",
        y_log=True,
        style="markers_lines",
    )
    plt_ibci = Plot(
        r"$IBCI(V_{\mathrm{BE}})$",
        x_specifier=specifiers.VOLTAGE + "B" + "E",
        y_specifier=specifiers.CURRENT + "B",
        y_log=True,
        style="markers_lines",
    )
    plt_it = Plot(
        r"IT(VBE)",
        x_specifier=specifiers.VOLTAGE + "B" + "E",
        y_specifier=specifiers.CURRENT + "C",
        y_log=True,
        style="markers_lines",
    )
    plt_temp = Plot(
        r"$dT(VBE)$",
        x_specifier=specifiers.VOLTAGE + "B" + "E",
        y_specifier=specifiers.TEMPERATURE,
        style="markers_lines",
    )
    plt_ft = Plot(
        r"FT(VBE)",
        x_specifier=specifiers.CURRENT + "C",
        y_specifier=specifiers.TRANSIT_FREQUENCY,
        x_log=True,
        style="markers_lines",
    )
    plt_vbici = Plot(
        r"VBICI(VBE)",
        x_specifier=specifiers.VOLTAGE + "B",
        y_specifier=specifiers.VOLTAGE + "Bi" + "Ci",
        style="markers_lines",
    )
    plt_vbiei = Plot(
        r"VBIEI(VBE)",
        x_specifier=specifiers.VOLTAGE + "B",
        y_specifier=specifiers.VOLTAGE + "Bi" + "Ei",
        style="markers_lines",
    )
    # end define plots

    # fill plots
    # gummel
    plt_gummel.add_data_set(data_ngspice["V_B"], data_ngspice["I_C"], label="ngspice $I_C$")
    plt_gummel.add_data_set(data_ads["V_B"], data_ads["I_C"], label="ADS $I_C$")
    plt_gummel.add_data_set(
        data_ngspice["V_B"],
        data_ngspice["I_B"],  # IB from ngspice short is wrong
        # data_ngspice['VBBP']/mc['rbx'].value,
        label="ngspice $I_B$",
    )
    # plt_gummel.add_data_set(
    #     data_ngspice['V_B'],
    #     data_ngspice['IBEI'], #IB from ngspice short is wrong
    #     #data_ngspice['VBBP']/mc['rbx'].value,
    #     label='ngspice $IBEI$'
    # )
    plt_gummel.add_data_set(data_ads["V_B"], data_ads["I_B"], label="ADS $I_B$")

    # ibei
    plt_ibei.add_data_set(data_ngspice["V_B"], data_ngspice["IBEI"], label="ngspice ibei")
    plt_ibei.add_data_set(data_ads["V_B"], data_ads["IBIEI"], label="ADS ibei")
    # ibci
    plt_ibci.add_data_set(data_ngspice["V_B"], data_ngspice["IBCI"], label="ngspice ibci")
    plt_ibci.add_data_set(data_ads["V_B"], data_ads["IBICI"], label="ADS ibci")
    # it
    plt_it.add_data_set(data_ngspice["V_B"], data_ngspice["IT"], label="ngspice it")
    plt_it.add_data_set(data_ads["V_B"], data_ads["IT_"], label="ADS it")
    # temp
    plt_temp.add_data_set(data_ngspice["V_B"], data_ngspice["V_T"], label="ngspice vt")
    plt_temp.add_data_set(data_ads["V_B"], data_ads["DTSH"], label="ADS it")
    # ft
    try:
        plt_ft.add_data_set(
            data_ngspice["I_C"], data_ngspice[specifiers.TRANSIT_FREQUENCY], label="ngspice "
        )
        plt_ft.add_data_set(data_ads["I_C"], data_ads[specifiers.TRANSIT_FREQUENCY], label="ADS ")
    except KeyError:
        pass
    # vbici
    plt_vbici.add_data_set(data_ngspice["V_B"], data_ngspice["VBICI"], label="ngspice ")
    plt_vbici.add_data_set(data_ads["V_B"], data_ads["_V_BiE"] - data_ads["_V_CiE"], label="ADS ")
    # vbiei
    plt_vbiei.add_data_set(data_ngspice["V_B"], data_ngspice["VBIEI"], label="ngspice ")
    plt_vbiei.add_data_set(data_ads["V_B"], data_ads["_V_BiE"] - data_ads["_V_EiE"], label="ADS ")

    plt_vbiei.plot_py(show=False)
    plt_vbici.plot_py(show=False)
    plt_ft.plot_py(show=False)
    plt_temp.plot_py(show=False)
    plt_it.plot_py(show=False)
    plt_ibci.plot_py(show=False)
    plt_ibei.plot_py(show=False)
    plt_gummel.plot_py(show=True)



""" test ngspice vs ADS using a real modelcard in a systematic way.
"""
import logging
from pathlib import Path
from DMT.core import SimCon, Plot, DutType, Sweep, specifiers, McParameter, sub_specifiers
from DMT.ngspice import DutNgspice
from DMT.ADS import DutAds
from DMT.hl2 import McHicum, VA_FILES

import numpy as np
import copy

folder_path = Path(__file__).resolve().parent
test_path = folder_path.parent

logging.basicConfig(
    level=logging.DEBUG,
    format="%(levelname)s - %(message)s",
    filename=folder_path.parent.parent / "logs" / "test_ADS_nqs.log",
    filemode="w",
)


def compare_ngspice_ads_real_mc(mc):
    rth_de = McParameter(name="rth_de", value=0)
    try:
        mc.add(rth_de)
    except:
        pass

    dut_ads = DutAds(
        None, DutType.npn, mc, nodes="C,B,E,S,T", reference_node="E"
    )  # ADS with VA file

    # create a sweep with ONE operating point
    sweepdef = [
        {
            "var_name": specifiers.FREQUENCY,
            "sweep_order": 4,
            "sweep_type": "LOG",
            "value_def": [9, 12, 101],
        },
        {
            "var_name": specifiers.VOLTAGE + "B",
            "sweep_order": 3,
            "sweep_type": "LIN",
            "value_def": [0.7, 1.1, 41],
        },
        {
            "var_name": specifiers.VOLTAGE + "C",
            "sweep_order": 2,
            "sweep_type": "CON",
            "value_def": [2],
        },
        {
            "var_name": specifiers.VOLTAGE + "E",
            "sweep_order": 1,
            "sweep_type": "CON",
            "value_def": [0],
        },
    ]
    outputdef = ["I_C", "I_B"]
    othervar = {specifiers.TEMPERATURE: 298}
    sweep = Sweep("gummel", sweepdef=sweepdef, outputdef=outputdef, othervar=othervar)

    # first simulate ADS
    sim_con = SimCon(n_core=1, t_max=30)
    sim_con.append_simulation(dut=dut_ads, sweep=sweep)
    sim_con.run_and_read(force=False)

    # read initial conditions
    data_ads = dut_ads.get_data(sweep=sweep)
    ve = 0
    vb = data_ads["VBE"].to_numpy()[0]
    vc = data_ads["VCE"].to_numpy()[0]
    vbi = data_ads["VBiE"].to_numpy()[0]
    vbp = data_ads["VBpE"].to_numpy()[0]
    vci = data_ads["VCiE"].to_numpy()[0]
    vei = data_ads["VEiE"].to_numpy()[0]
    vt = data_ads["DTSH"].to_numpy()[0]
    delta = 1
    initial_conditions = {
        "icVE": ve * delta,
        "icVEi": vei * delta,
        "icVC": vc * delta,
        "icVCi": vci * delta,
        "icVB": vb * delta,
        "icVBi": vbi * delta,
        "icVBp": vbp * delta,
        "icVt": vt * delta,
    }

    # get rth stuff from ads
    rth_t = data_ads["rth_t"].to_numpy()
    pterm = data_ads["pterm"].to_numpy()
    dT = rth_t * pterm

    assert np.isclose(vt, dT[0], rtol=1e-3)

    # simulate ngspice using initial condition from ADS
    dut_ngspice = DutNgspice(
        None,
        DutType.npn,
        mc,
        nodes="C,B,E,S,T",
        reference_node="E",
        copy_va_files=False,  # ngspice without VA file! -> use internal
        simulator_command="ngspice",
        initial_conditions=initial_conditions,
        simulator_options={"GMIN": 1e-18, "ABSTOL": 1e-15},
    )
    sim_con.append_simulation(dut=dut_ngspice, sweep=sweep)
    sim_con.run_and_read(force=True)

    data_ngspice = dut_ngspice.get_data(sweep=sweep)

    # #ngspice operating point
    # vbiei  = data_ngspice['VBIEI'].to_numpy()
    # vbici  = data_ngspice['VBICI'].to_numpy()
    # vciei  = data_ngspice['VCIEI'].to_numpy()
    # temp   = data_ngspice['TK'].to_numpy()
    # #ngspice currents
    # ibei_ngspice = data_ngspice['IBEI'].to_numpy()
    # ibei_ads     = data_ads['IBIEI'].to_numpy()
    # it_ngspice   = data_ngspice['IT'].to_numpy()
    # it_ads       = data_ads['IT_'].to_numpy()

    # ensure some stuff
    ensure = [
        specifiers.TRANSIT_FREQUENCY,
        specifiers.CAPACITANCE + "B" + "E",
        specifiers.CAPACITANCE + "B" + "C",
        specifiers.CAPACITANCE + "C" + "E",
        specifiers.SS_PARA_Y + "B" + "B" + sub_specifiers.IMAG,
        specifiers.SS_PARA_Y + "B" + "B" + sub_specifiers.REAL,
        specifiers.SS_PARA_Y + "C" + "B" + sub_specifiers.REAL,
        specifiers.SS_PARA_Y + "C" + "B" + sub_specifiers.IMAG,
        specifiers.SS_PARA_Y + "B" + "C" + sub_specifiers.REAL,
        specifiers.SS_PARA_Y + "B" + "C" + sub_specifiers.IMAG,
        specifiers.SS_PARA_Y + "C" + "C" + sub_specifiers.REAL,
        specifiers.SS_PARA_Y + "C" + "C" + sub_specifiers.IMAG,
    ]
    for speci in ensure:
        data_ngspice.ensure_specifier_column(speci, ports=["B", "C"])
        data_ads.ensure_specifier_column(speci, ports=["B", "C"])

    for col in data_ngspice.columns:
        col_data = data_ngspice[col].to_numpy()
        if np.all(np.imag(col_data) == 0):
            data_ngspice[col] = data_ngspice[col].astype(float)

    return data_ngspice, data_ads


if __name__ == "__main__":
    show = False  # If True, plot in maptplotlib, else generate Tex

    # load a modelcard
    mc = McHicum(load_model_from_path=folder_path / "ngspice_example_modelcard.lib")

    mc = mc.get_clean_modelcard()
    mc.set_values(
        {
            # 'flnqs' : 0,
            # 'flsh'  : 1,
            "cbcpar": 0,
            "cjcx0": 0,
            # 'fqi'    : 0,
            # 'fcrbi'  : 0,
            # 'cjci0'  : 0,
            # 'ibcxs'  : 0,
            # 'csu'  : 0,
            #'rsu'  : 100,
            "iscs": 0,
            "itss": 0,
        },
        force=True,
    )
    mc.va_file = VA_FILES["L2V2.4.0_internal"]
    # end load modelcard

    # define plots
    plt_gummel = Plot(
        r"$I(V_{\mathrm{BE}})$",
        x_specifier=specifiers.VOLTAGE + "B" + "E",
        y_specifier=specifiers.CURRENT + "C",
        y_log=True,
        style="comparison_4",
        legend_location="upper right outer",
    )
    plt_ft = Plot(
        r"FT(VBE)",
        x_specifier=specifiers.CURRENT + "C",
        y_specifier=specifiers.TRANSIT_FREQUENCY,
        x_log=True,
        y_scale=1e-9,
        style="xtraction_color",
        legend_location="upper right outer",
    )
    plt_ft.x_limits = (1e-5, 1e0)
    plt_ft.y_limits = (0, 600)
    plt_rey21 = Plot(
        r"ReY21(f)",
        x_specifier=specifiers.FREQUENCY,
        y_specifier=specifiers.SS_PARA_Y + "2" + "1" + sub_specifiers.REAL,
        x_log=True,
        style="xtraction_color",
        x_scale=1e-9,
        legend_location="upper right outer",
    )
    plt_imy21 = Plot(
        r"ImY21(f)",
        x_specifier=specifiers.FREQUENCY,
        y_specifier=specifiers.SS_PARA_Y + "2" + "1" + sub_specifiers.IMAG,
        x_log=True,
        style="xtraction_color",
        x_scale=1e-9,
        legend_location="upper right outer",
    )
    plt_magy21 = Plot(
        r"MagY21(f)",
        x_specifier=specifiers.FREQUENCY,
        y_label="mag(Y21)",
        x_log=True,
        style="xtraction_color",
        x_scale=1e-9,
        legend_location="upper right outer",
    )
    plt_phasey21 = Plot(
        r"PhaseY21(f)",
        x_specifier=specifiers.FREQUENCY,
        y_label=r"Phase (degree)",
        x_log=True,
        style="xtraction_color",
        x_scale=1e-9,
        legend_location="upper right outer",
    )

    mc_no_nqs = copy.deepcopy(mc)
    mc_nqs = copy.deepcopy(mc)
    mc_nqs.set_values({"flnqs": 1, "alqf": 0, "alit": 0})
    mcs = []
    for val in np.linspace(0, 1, 6):
        mc = copy.deepcopy(mc_nqs)
        mc.set_values({"alqf": val})
        mc.set_values({"alit": 0})
        mcs.append(mc)

    for val in np.linspace(0, 1, 6):
        mc = copy.deepcopy(mc_nqs)
        mc.set_values({"alqf": 1})
        mc.set_values({"alit": val})
        mcs.append(mc)

    for mc in mcs:
        # if mc['flnqs'].value == 1:
        #     label_add = 'flnqs=1'
        # else:
        #     label_add = 'flnqs=0'

        label_add = "alqf={0:1.1f}, alit={1:1.1f}".format(mc["alqf"].value, mc["alit"].value)

        data_ngspice, data_ads = compare_ngspice_ads_real_mc(mc)

        # fill plots
        data_ngspice_10GHz = data_ngspice[np.isclose(data_ngspice[specifiers.FREQUENCY], 1e9)]
        data_ads_10GHz = data_ads[np.isclose(data_ads[specifiers.FREQUENCY], 1e9)]
        # gummel
        plt_gummel.add_data_set(
            data_ngspice_10GHz["V_B"], data_ngspice_10GHz["I_C"], label="NGS " + label_add
        )
        plt_gummel.add_data_set(
            data_ads_10GHz["V_B"], data_ads_10GHz["I_C"], label="ADS " + label_add
        )
        plt_gummel.add_data_set(
            data_ngspice_10GHz["V_B"], data_ngspice_10GHz["I_B"]  # IB from ngspice short is wrong
        )
        plt_gummel.add_data_set(data_ads_10GHz["V_B"], data_ads_10GHz["I_B"])

        plt_ft.add_data_set(
            data_ngspice_10GHz["I_C"],
            data_ngspice_10GHz[specifiers.TRANSIT_FREQUENCY],
            label="NGS " + label_add,
        )
        plt_ft.add_data_set(
            data_ads_10GHz["I_C"],
            data_ads_10GHz[specifiers.TRANSIT_FREQUENCY],
            label="ADS " + label_add,
        )

        # find peak ft
        ft = data_ads_10GHz[specifiers.TRANSIT_FREQUENCY].to_numpy()
        vbe = data_ads_10GHz[specifiers.VOLTAGE + "B"].to_numpy()
        index_peak_ft = np.argmax(ft) + 2
        vbe_at_peak = vbe[index_peak_ft]
        data_ads = data_ads[
            np.isclose(data_ads[specifiers.VOLTAGE + "B" + sub_specifiers.FORCED], vbe_at_peak)
        ]
        data_ngs = data_ngspice[
            np.isclose(data_ngspice[specifiers.VOLTAGE + "B" + sub_specifiers.FORCED], vbe_at_peak)
        ]
        plt_rey21.add_data_set(
            data_ads[specifiers.FREQUENCY],
            -data_ads[specifiers.SS_PARA_Y + "B" + "C" + sub_specifiers.REAL],
            label="ads " + label_add,
        )
        plt_rey21.add_data_set(
            data_ngs[specifiers.FREQUENCY],
            -data_ngs[specifiers.SS_PARA_Y + "B" + "C" + sub_specifiers.REAL],
            label="ngs " + label_add,
        )
        plt_imy21.add_data_set(
            data_ads[specifiers.FREQUENCY],
            -data_ads[specifiers.SS_PARA_Y + "B" + "C" + sub_specifiers.IMAG],
            label="ads " + label_add,
        )
        plt_imy21.add_data_set(
            data_ngs[specifiers.FREQUENCY],
            -data_ngs[specifiers.SS_PARA_Y + "B" + "C" + sub_specifiers.IMAG],
            label="ngs " + label_add,
        )
        plt_magy21.add_data_set(
            data_ads[specifiers.FREQUENCY],
            np.abs(data_ads[specifiers.SS_PARA_Y + "B" + "C"]),
            label="ads " + label_add,
        )
        plt_magy21.add_data_set(
            data_ngs[specifiers.FREQUENCY],
            np.abs(data_ngs[specifiers.SS_PARA_Y + "B" + "C"]),
            label="ngs " + label_add,
        )
        plt_phasey21.add_data_set(
            data_ads[specifiers.FREQUENCY],
            -np.angle(data_ads[specifiers.SS_PARA_Y + "B" + "C"], deg=True),
            label="ads " + label_add,
        )
        plt_phasey21.add_data_set(
            data_ngs[specifiers.FREQUENCY],
            -np.angle(data_ngs[specifiers.SS_PARA_Y + "B" + "C"], deg=True),
            label="ngs " + label_add,
        )

    plts = [
        plt_ft,
        plt_gummel,
        plt_rey21,
        plt_imy21,
        plt_magy21,
        plt_phasey21,
    ]
    if show:
        for plt in plts[:-1]:
            plt.plot_py(show=False)

        plts[-1].plot_py(show=True)

    else:
        ###PRINT TO DOCUMENT FROM HERE#####################
        from pylatex import (
            Section,
            Subsection,
            Tabular,
            Math,
            Plot,
            Figure,
            Matrix,
            Alignat,
            Package,
            Itemize,
            NoEscape,
            Document,
        )
        from pylatex.base_classes import Arguments
        from pylatexenc.latex2text import LatexNodes2Text

        # from DMT.gui.utils import SubFile, CommandRefRange, CommandRef, CommandInput, CommandLabel from DMT.gui.utils import SubFile, CommandRefRange, CommandRef, CommandInput, CommandLabel from pylatexenc.latex2text import LatexNodes2Text
        from DMT.external import SubFile, CommandRef, CommandInput, CommandLabel

        width = "4.5in"

        doc = Document()
        fig_dir = folder_path / "documentation" / "figs"

        with doc.create(Section("Comparison ADS/NGspice")):
            doc.append(
                NoEscape(
                    r"In this document the ngspice HICUM/L2.4.3 model in ngspice is compared against ADS simulations."
                )
            )
            doc.append(NoEscape(r"The modelcard is taken from the ngspice examples directory. "))
            doc.append(
                NoEscape(
                    r"This document is auto-generated using Pylatex. The shown ft are calculated from simulated Y-parameters. "
                )
            )
            # doc.append(NoEscape(r'\FloatBarrier'))
            with doc.create(Subsection("Example Modelcard Characteristics")):
                doc.append(
                    NoEscape(
                        r"First, the transfer characteristics and transit frequency as a function of Jc are shown: "
                    )
                )
                for plt in [plt_ft, plt_gummel]:
                    plt_name = plt.save_tikz(
                        fig_dir, width=width, mark_repeat=2, legend_location="upper right outer"
                    )
                    with doc.create(Figure(position="h!")) as _plot:
                        _plot.append(
                            CommandInput(arguments=Arguments('"' + str(fig_dir / plt_name) + '"'))
                        )
                        _plot.add_caption(NoEscape(plt.name))
                        _plot.append(
                            CommandLabel(
                                arguments=Arguments(LatexNodes2Text().latex_to_text(plt.name))
                            )
                        )
            doc.append(NoEscape(r"\newpage"))
            with doc.create(Subsection("Frequency Dependence Y21")):
                doc.append(
                    NoEscape(r"Next, the frequency dependence of Y21 at peak ft are shown: ")
                )
                for plt in [plt_rey21, plt_imy21, plt_phasey21, plt_magy21]:
                    plt_name = plt.save_tikz(
                        fig_dir, width=width, mark_repeat=2, legend_location="upper right outer"
                    )
                    with doc.create(Figure(position="h!")) as _plot:
                        _plot.append(
                            CommandInput(arguments=Arguments('"' + str(fig_dir / plt_name) + '"'))
                        )
                        _plot.add_caption(NoEscape(plt.name))
                        _plot.append(
                            CommandLabel(
                                arguments=Arguments(LatexNodes2Text().latex_to_text(plt.name))
                            )
                        )

        doc.generate_tex(folder_path / "documentation" / "comparison_ngspice_ads_nqs")



""" test ngspice input file generation.
"""
import types
import logging
from pathlib import Path
from DMT.core import DutType, Sweep, specifiers, SimCon, Plot, MCard
from DMT.core.circuit import Circuit, CircuitElement, RESISTANCE, CAPACITANCE, SHORT, VOLTAGE

from DMT.ngspice import DutNgspice
import numpy as np

folder_path = Path(__file__).resolve().parent

logging.basicConfig(
    level=logging.DEBUG,
    format="%(levelname)s - %(message)s",
    filename=folder_path.parent.parent / "logs" / "test_ngspice.log",
    filemode="w",
)


def get_circuit(self, use_build_in=False, topology="common_emitter", **kwargs):
    """

    Parameter
    ------------
    circuit_type : str
        For allowed types, see above
    modelcard : :class:`~DMT.hl2.mc_hicum.McHicum`

    Returns
    -------
    circuit : :class:`~DMT.core.circuit.Circuit`

    """
    node_emitter = next(f"n_{node.upper()}" for node in self.nodes_list if "E" in node.upper())
    node_base = next(f"n_{node.upper()}" for node in self.nodes_list if "B" in node.upper())
    node_collector = next(f"n_{node.upper()}" for node in self.nodes_list if "C" in node.upper())
    node_substrate = next(f"n_{node.upper()}" for node in self.nodes_list if "S" in node.upper())
    node_temperature = next(f"n_{node.upper()}" for node in self.nodes_list if "T" in node.upper())

    circuit_elements = []
    # model instance
    circuit_elements.append(
        CircuitElement(
            self.default_module_name,
            "Q_H",
            [f"n_{node.upper()}" for node in self.nodes_list],
            # ["n_C", "n_B", "n_E"],
            parameters=self,
        )
    )

    if topology == "common_emitter":

        # BASE NODE CONNECTION #############
        # metal resistance between contact base point and real collector
        try:
            rbm = self.get("_rbm").value
        except KeyError:
            rbm = 1e-3

        circuit_elements.append(
            CircuitElement(
                RESISTANCE, "Rbm", ["n_B_FORCED", node_base], parameters=[("R", str(rbm))]
            )
        )
        # shorts for current measurement
        circuit_elements.append(
            CircuitElement(
                SHORT,
                "I_B",
                ["n_BX", "n_B_FORCED"],
            )
        )
        # capacitance since AC already deembeded Rbm
        circuit_elements.append(
            CircuitElement(
                CAPACITANCE, "Cbm", ["n_B_FORCED", node_base], parameters=[("C", str(1))]
            )
        )

        # COLLECTOR NODE CONNECTION #############
        circuit_elements.append(
            CircuitElement(
                SHORT,
                "I_C",
                ["n_CX", "n_C_FORCED"],
            )
        )
        # metal resistance between contact collector point and real collector
        try:
            rcm = self.get("_rcm").value
        except KeyError:
            rcm = 1e-3

        circuit_elements.append(
            CircuitElement(
                RESISTANCE, "Rcm", ["n_C_FORCED", node_collector], parameters=[("R", str(rcm))]
            )
        )
        # capacitance since AC already deembeded Rcm
        circuit_elements.append(
            CircuitElement(
                CAPACITANCE, "Ccm", ["n_C_FORCED", node_collector], parameters=[("C", str(1))]
            )
        )
        # EMITTER NODE CONNECTION #############
        circuit_elements.append(
            CircuitElement(
                SHORT,
                "I_E",
                ["n_EX", "n_E_FORCED"],
            )
        )
        # metal resistance between contact emiter point and real emiter
        try:
            rem = self.get("_rem").value
        except KeyError:
            rem = 1e-3

        circuit_elements.append(
            CircuitElement(
                RESISTANCE, "Rem", ["n_E_FORCED", node_emitter], parameters=[("R", str(rem))]
            )
        )
        # capacitance since AC already deembeded Rcm
        circuit_elements.append(
            CircuitElement(
                CAPACITANCE, "Cem", ["n_E_FORCED", node_emitter], parameters=[("C", str(1))]
            )
        )
        # add sources and thermal resistance
        circuit_elements.append(
            CircuitElement(VOLTAGE, "V_B", ["n_BX", "0"], parameters=[("Vdc", "V_B"), ("Vac", "1")])
        )
        circuit_elements.append(
            CircuitElement(VOLTAGE, "V_C", ["n_CX", "0"], parameters=[("Vdc", "V_C"), ("Vac", "1")])
        )
        circuit_elements.append(
            CircuitElement(VOLTAGE, "V_E", ["n_EX", "0"], parameters=[("Vdc", "V_E"), ("Vac", "1")])
        )

        # metal resistance between contact emitter potential and substrate contact
        if len(self.nodes_list) > 3:
            circuit_elements.append(
                CircuitElement(
                    SHORT,
                    "I_S",
                    ["n_SX", node_substrate],
                )
            )
            try:
                rsm = self.get("_rsm").value
            except KeyError:
                rsm = 5
            circuit_elements.append(
                CircuitElement(RESISTANCE, "R_S", ["n_SX", "n_EX"], parameters=[("R", str(rsm))])
            )
        if len(self.nodes_list) > 4:
            circuit_elements.append(
                CircuitElement(
                    RESISTANCE, "R_t", [node_temperature, "0"], parameters=[("R", "1e9")]
                )
            )
        circuit_elements += [
            "V_B=0",
            "V_C=0",
            "V_S=0",
            "V_E=0",
            "ac_switch=0",
            "V_B_ac=1-ac_switch",
            "V_C_ac=ac_switch",
            "V_S_ac=0",
            "V_E_ac=0",
        ]
    else:
        raise IOError("The circuit type " + topology + " is unknown!")

    return Circuit(circuit_elements)


def test_ngspice():
    col_vb = specifiers.VOLTAGE + "B"
    col_ib = specifiers.CURRENT + "B"
    col_ic = specifiers.CURRENT + "C"

    mc_D21 = MCard(
        ["C", "B", "E", "S", "T"],
        default_module_name="",
        default_subckt_name="",
        va_file=folder_path.parent
        / "test_core_no_interfaces"
        / "test_va_code"
        / "hicuml2"
        / "hicumL2V2p4p0_release.va",
    )
    mc_D21.load_model_parameters(
        folder_path.parent
        / "test_core_no_interfaces"
        / "test_modelcards"
        / "IHP_ECE704_03_para_D21.mat",
    )
    mc_D21.update_from_vae(remove_old_parameters=True)
    mc_D21.get_circuit = types.MethodType(get_circuit, mc_D21)

    dut = DutNgspice(None, DutType.npn, mc_D21, nodes="C,B,E,S,T", reference_node="E")
    # dut = DutXyce(None, DutType.npn, mc_D21, nodes="C,B,E,S,T", reference_node="E")
    duts = [dut]

    # create a sweep
    sweepdef = [
        {
            "var_name": specifiers.FREQUENCY,
            "sweep_order": 4,
            "sweep_type": "LOG",
            "value_def": [1, 2, 11],
        },
        {
            "var_name": specifiers.VOLTAGE + "B",
            "sweep_order": 3,
            "sweep_type": "LIN",
            "value_def": [0.5, 1, 11],
        },
        {
            "var_name": specifiers.VOLTAGE + "C",
            "sweep_order": 2,
            "sweep_type": "CON",
            "value_def": [1],
        },
        {
            "var_name": specifiers.VOLTAGE + "E",
            "sweep_order": 1,
            "sweep_type": "CON",
            "value_def": [0],
        },
    ]
    outputdef = ["I_C", "I_B"]
    othervar = {"TEMP": 300}
    sweep = Sweep("gummel", sweepdef=sweepdef, outputdef=outputdef, othervar=othervar)

    sim_con = SimCon()

    sim_con.append_simulation(dut=dut, sweep=sweep)
    sim_con.run_and_read(force=True, remove_simulations=False)

    df = dut.get_data(sweep=sweep)
    df = df[np.isclose(df[specifiers.FREQUENCY], 100)]
    vb = np.real(df[col_vb].to_numpy())
    ic = np.real(df[col_ic].to_numpy())
    ib = np.real(df[col_ib].to_numpy())

    assert np.isclose(
        vb,
        np.array(
            [
                0.5,
                0.55,
                0.6,
                0.65,
                0.7,
                0.75,
                0.799999998,
                0.849999988,
                0.899999936,
                0.949999737,
                0.999998983,
            ]
        ),
    ).all()
    assert np.isclose(
        ic,
        np.array(
            [
                1.14418981e-08,
                7.72463409e-08,
                5.2127632e-07,
                3.51490644e-06,
                2.36611728e-05,
                0.000158456473,
                0.00102887822,
                0.00566126767,
                0.0207667184,
                0.0484566543,
                0.06658047,
            ]
        ),
    ).all()
    assert np.isclose(
        ib,
        np.array(
            [
                3.81987775e-11,
                1.86560101e-10,
                1.10651399e-09,
                7.04358172e-09,
                4.57921487e-08,
                2.99428393e-07,
                1.95696248e-06,
                1.22903099e-05,
                6.44306933e-05,
                0.000262592983,
                0.00101685885,
            ]
        ),
    ).all()

    return duts, sweep


if __name__ == "__main__":
    duts, sweep = test_ngspice()

    col_vbc = specifiers.VOLTAGE + ["B", "C"]
    col_vbe = specifiers.VOLTAGE + ["B", "E"]
    col_vce = specifiers.VOLTAGE + ["C", "E"]
    col_ib = specifiers.CURRENT + "B"
    col_ic = specifiers.CURRENT + "C"

    plt_gummel = Plot(
        r"$I_{\mathrm{C}}(V_{\mathrm{BE}})$",
        x_label=r"$V_{\mathrm{BE}}$",
        y_label=r"$I_{\mathrm{C}}$",
        y_log=True,
        style="bw",
    )
    plt_ib = Plot(
        r"$I_{\mathrm{B}}(V_{\mathrm{BE}})$",
        x_label=r"$V_{\mathrm{BE}}$",
        y_label=r"$I_{\mathrm{B}}$",
        y_log=True,
        style="bw",
    )

    for dut in duts:
        dut_name = dut.name[5:7]
        df = dut.get_data(sweep=sweep)

        df.ensure_specifier_column(col_vbe, ports=dut.nodes)
        df.ensure_specifier_column(col_vbc, ports=dut.nodes)
        df.ensure_specifier_column(col_vce, ports=dut.nodes)

        for _index, vce, data in df.iter_unique_col(col_vce, decimals=3):
            data = data[np.isclose(data[specifiers.FREQUENCY], 100)]

            plt_gummel.add_data_set(
                np.real(data[col_vbe].to_numpy()),
                np.real(data[col_ic].to_numpy()),
                label=dut_name + " $V_{{CE}} = {0:1.2f} V$".format(np.real(vce)),
            )
            plt_ib.add_data_set(
                np.real(data[col_vbe].to_numpy()),
                np.real(data[col_ib].to_numpy()),
                label=dut_name + " $V_{{CE}} = {0:1.2f} V$".format(np.real(vce)),
            )

    plt_ib.plot_pyqtgraph(show=False)
    plt_gummel.plot_pyqtgraph(show=True)
    # plt_ib.plot_py(show=False, use_tex=True)
    # plt_gummel.plot_py(show=True, use_tex=False)



""" OS interaction of DMT
"""
# DMT_core
# Copyright (C) from 2022  SemiMod
# Copyright (C) until 2021  Markus Müller, Mario Krattenmacher and Pascal Kuthe
# <https://gitlab.com/dmt-development/dmt-core>
#
# This file is part of DMT_core.
#
# DMT_core is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DMT_core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>
import os
import shutil
import re
from pathlib import Path

##MARKUS CONTEXT MANAGER AWESOMENESS ###############
# this is a cd command that support context_manager python commands.


class cd:
    """Context manager for changing the current working directory

    Usage::

        with cd(dir):
            pass

    """

    def __init__(self, newPath):
        self.savedPath = ""
        self.newPath = os.path.expanduser(newPath)

    def __enter__(self):
        self.savedPath = os.getcwd()
        os.chdir(self.newPath)

    def __exit__(self, etype, value, traceback):
        os.chdir(self.savedPath)


def recursive_copy(src, dst, force=False):
    """Recursively copies a full directory to a new destination folder.

    Parameters
    ----------
    src : str or os.Pathlike
        Path to the source folder
    dst : str or os.Pathlike
        Path to the destination folder
    force : bool, optional
        When true, all files are overwritten. Else, already existing ones are kept.
    """
    if not isinstance(src, Path):
        src = Path(src)
    if not isinstance(dst, Path):
        dst = Path(dst)
    for item_src in src.iterdir():
        item_dst = dst / item_src.name
        if item_src.is_file():
            if not item_dst.exists() or force:
                shutil.copy(item_src, dst)

        elif item_src.is_dir():
            item_dst.mkdir(exist_ok=True)
            recursive_copy(item_src, item_dst)
        else:
            raise ValueError("DMT->recursive_copy: I do not know this filettype.")


def rmtree(root):
    """rmtree method for Path objects

    Parameters
    ----------
    root : str or os.Pathlike
        Directory to remove
    """
    if not isinstance(root, Path):
        root = Path(root)
    if not root.exists():
        return  # nothing to do here -.-

    for p in root.iterdir():
        if p.is_dir():
            rmtree(p)
        else:
            p.unlink()

    root.rmdir()


def slugify(s: str) -> str:
    """https://stackoverflow.com/questions/295135/turn-a-string-into-a-valid-filename
    Normalizes string, converts to lowercase, removes non-alpha characters,
    and converts spaces to hyphens.
    """
    s = str(s).strip().replace(" ", "_")
    s = s.replace(".", "_dot_")
    s = s.replace("@", "_at_")
    return re.sub(r"(?u)[^-\w.]", "", s)



""" test ADS simulation and plotting of output data.
"""
import time
from pathlib import Path
import logging
import numpy as np
from DMT.core import SimCon, specifiers, DutType, Sweep, Plot, sub_specifiers
from DMT.hl2 import McHicum, VA_FILES
from DMT.ngspice import DutNgspice
from DMT.ADS import DutAds

sim_ngspice = True
sim_ads = True
show = True
force = False

folder_path = Path(__file__).resolve().parent
test_path = folder_path.parent

logging.basicConfig(
    level=logging.DEBUG,
    format="%(levelname)s - %(message)s",
    filename=folder_path.parent.parent / "logs" / "test_ADS_many_sweeps.log",
    filemode="w",
)


def compare_ngspice_many_sweeps():
    # get HICUM model
    mc_D21 = McHicum(
        va_file=test_path
        / "test_core_no_interfaces"
        / "test_va_code"
        / "hicuml2"
        / "hicumL2V2p4p0_release.va",
        load_model_from_path=test_path
        / "test_core_no_interfaces"
        / "test_modelcards"
        / "IHP_ECE704_03_para_D21.mat",
    )
    mc_D21.va_file = VA_FILES["L2V2.4.0_internal"]
    mc_D21.remove(["betcbar", "ibexs", "mbex", "rx", "version"])

    # turn off (on) self heating for the moment
    flsh = mc_D21["flsh"]
    flsh.value = 1
    mc_D21.set(flsh)
    re = mc_D21["re"]
    re.value = 7
    mc_D21.set(re)
    rth = mc_D21["rth"]
    rth.value = 3e3
    mc_D21.set(rth)
    alrth = mc_D21["alrth"]
    alrth.value = 1e-4
    mc_D21.set(alrth)
    # zetare seems to be the source of evil

    # create DUTs
    dut_ngspice = DutNgspice(None, DutType.npn, mc_D21, nodes="C,B,E,S,T", reference_node="E")
    dut_ads = DutAds(None, DutType.npn, mc_D21, nodes="C,B,E,S,T", reference_node="E")

    # VBE :
    vbe_value_def = [0.7, 1.1, 81]  # from 0 to 1 V in 5 steps
    vbc_value_defs = np.linspace(0.5, -0.5, 5)  # from -0.5 to 0.5 V in 5 steps
    vbc_value_defs = [-0.4, 0, 0.4]

    # dmt for multiple simulations
    dmt = SimCon(t_max=20, n_core=1)

    # sweeps
    outputdef = ["I_C", "I_B"]
    othervar = {"TEMP": 300, "w": 10, "l": 0.25}

    sweeps = []
    for vc_ in vbc_value_defs:
        # create a sweep
        sweepdef = [
            {
                "var_name": specifiers.FREQUENCY,
                "sweep_order": 3,
                "sweep_type": "CONST",
                "value_def": [1e9],
            },
            {
                "var_name": specifiers.VOLTAGE + "B",
                "sweep_order": 2,
                "sweep_type": "LIN",
                "value_def": vbe_value_def,
            },
            {
                "var_name": specifiers.VOLTAGE + "C",
                "sweep_order": 2,
                "sweep_type": "SYNC",
                "master": "V_B",
                "offset": vc_,
            },
            {
                "var_name": specifiers.VOLTAGE + "E",
                "sweep_order": 1,
                "sweep_type": "CON",
                "value_def": [0],
            },
        ]
        sweeps.append(Sweep("gummel", sweepdef=sweepdef, outputdef=outputdef, othervar=othervar))

    time_ads = 0
    if sim_ads:
        for sweep in sweeps:
            dmt.append_simulation(dut_ads, sweep)

        start = time.time()
        dmt.run_and_read(force=force, remove_simulations=False)
        time_ads = time.time() - start

    time_ngspice = 0
    if sim_ngspice:
        for sweep in sweeps:
            dmt.append_simulation(dut_ngspice, sweep)

        start = time.time()
        dmt.run_and_read(force=force, remove_simulations=False)
        time_ngspice = time.time() - start

    return time_ads, time_ngspice, vbc_value_defs, dut_ads, dut_ngspice


if __name__ == "__main__":
    time_ads, time_ngspice, vbc_value_defs, dut_ads, dut_ngspice = compare_ngspice_many_sweeps()

    # now some plotting
    plt_gummel = Plot(
        "I_C(V_BE)",
        x_label=r"$V_{\mathrm{BE}}/\si{\volt}$",
        x_scale=1,
        y_label=r"$I_{\mathrm{C}}/\si{\milli\ampere}$",
        y_scale=1e3,
        y_log=True,
        legend_location="upper left",
        style="xtraction_color",
    )
    plt_rey21 = Plot(
        "ReY_21(V_BE)",
        x_label=r"$V_{\mathrm{BE}}/\si{\volt}$",
        x_scale=1,
        y_label=r"$\Re\left\{ Y_{\mathrm{21}} \right\}/\si{\milli\siemens}$",
        y_scale=1e3,
        y_log=True,
        legend_location="upper left",
        style="xtraction_color",
    )
    plt_imy21 = Plot(
        "ImY_21(V_BE)",
        x_label=r"$V_{\mathrm{BE}}/\si{\volt}$",
        x_scale=1,
        y_label=r"$\Im\left\{ Y_{\mathrm{21}} \right\}/\si{\milli\siemens}$",
        y_scale=1e3,
        y_log=True,
        legend_location="upper left",
        style="xtraction_color",
    )
    plt_y11 = Plot(
        "Y_11(V_BE)",
        x_label=r"$V_{\mathrm{BE}}/\si{\volt}$",
        x_scale=1,
        y_label=r"$\left| Y_{\mathrm{21}} \right|/\si{\milli\siemens}$",
        y_scale=1e3,
        y_log=True,
        legend_location="upper left",
        style="xtraction_color",
    )
    plt_ft = Plot(
        "f_T(I_C)",
        x_label=r"$I_{\mathrm{C}}/\si{\milli\ampere}$",
        x_scale=1e3,
        y_label=r"$f_{\mathrm{t}}/\si{\giga\hertz}$",
        y_scale=1e-9,
        y_log=False,
        x_log=True,
        legend_location="upper left",
        style="xtraction_color",
    )
    plts = [plt_gummel, plt_rey21, plt_imy21, plt_y11, plt_ft]

    if sim_ads and not sim_ngspice:
        for vbc, key_ads in zip(vbc_value_defs, dut_ads.data.keys()):
            data_ads = dut_ads.data[key_ads]
            data_ads = data_ads.calc_ft("B", "C")

            plt_gummel.add_data_set(
                np.asarray(data_ads["V_B"]),
                np.asarray(abs(data_ads["I_C"])),
                # label=r'$ads\,V_{{BC}} = \SI{' + str(vbc) + r'}{\volt}$',
            )
            plt_rey21.add_data_set(
                np.asarray(abs(data_ads[data_ads["FREQ"] == 1e9]["V_B"])),
                np.real(data_ads[data_ads["FREQ"] == 1e9]["Y_CB"]),
                # label=r'$ads V_{{BC}} = \SI{' + str(vbc) + r'}{\volt}$',
            )
            plt_imy21.add_data_set(
                np.asarray(abs(data_ads[data_ads["FREQ"] == 1e9]["V_B"])),
                np.imag(data_ads[data_ads["FREQ"] == 1e9]["Y_CB"]),
                # label=r'$ads V_{{BC}} = \SI{' + str(vbc) + r'}{\volt}$',
            )
            plt_y11.add_data_set(
                np.asarray(abs(data_ads[data_ads["FREQ"] == 1e9]["V_B"])),
                np.asarray(abs(data_ads[data_ads["FREQ"] == 1e9]["Y_BB"])),
                # label=r'$ads V_{{BC}} = \SI{' + str(vbc) + r'}{\volt}$',
            )
            plt_ft.add_data_set(
                np.asarray(abs(data_ads[data_ads["FREQ"] == 1e9]["I_C"])),
                np.asarray(abs(data_ads[data_ads["FREQ"] == 1e9][specifiers.TRANSIT_FREQUENCY])),
                # label=r'$ads V_{{BC}} = \SI{' + str(vbc) + r'}{\volt}$',
            )

    elif sim_ngspice and not sim_ads:
        for vbc, key_ngspice in zip(vbc_value_defs, dut_ngspice.data.keys()):
            data_ngspice = dut_ngspice.data[key_ngspice]
            data_ngspice = data_ngspice.calc_ft("B", "C")
            # data_simu    = dut_simu.data[key_simu]
            plt_gummel.add_data_set(
                np.asarray(abs(data_ngspice[data_ngspice["FREQ"] == 1e9]["V_B"])),
                np.asarray(abs(data_ngspice[data_ngspice["FREQ"] == 1e9]["I_C"])),
                label=r"$V_{{BC}} = \SI{" + str(vbc) + r"}{\volt}$",
            )
            plt_rey21.add_data_set(
                np.asarray(abs(data_ngspice[data_ngspice["FREQ"] == 1e9]["V_B"])),
                np.real(data_ngspice[data_ngspice["FREQ"] == 1e9]["Y_CB"]),
                label=r"$V_{{BC}} = \SI{" + str(vbc) + r"}{\volt}$",
            )
            plt_imy21.add_data_set(
                np.asarray(abs(data_ngspice[data_ngspice["FREQ"] == 1e9]["V_B"])),
                np.imag(data_ngspice[data_ngspice["FREQ"] == 1e9]["Y_CB"]),
                label=r"$V_{{BC}} = \SI{" + str(vbc) + r"}{\volt}$",
            )
            plt_y11.add_data_set(
                np.asarray(abs(data_ngspice[data_ngspice["FREQ"] == 1e9]["V_B"])),
                np.asarray(abs(data_ngspice[data_ngspice["FREQ"] == 1e9]["Y_BB"])),
                label=r"$V_{{BC}} = \SI{" + str(vbc) + r"}{\volt}$",
            )
            plt_ft.add_data_set(
                np.asarray(abs(data_ngspice[data_ngspice["FREQ"] == 1e9]["I_C"])),
                np.asarray(
                    abs(data_ngspice[data_ngspice["FREQ"] == 1e9][specifiers.TRANSIT_FREQUENCY])
                ),
                label=r"$V_{{BC}} = \SI{" + str(vbc) + r"}{\volt}$",
            )

    elif sim_ngspice and sim_ads:
        for vbc, key_ngspice, key_ads in zip(
            vbc_value_defs, dut_ngspice.data.keys(), dut_ads.data.keys()
        ):
            data_ngspice = dut_ngspice.data[key_ngspice]
            data_ngspice = data_ngspice.calc_ft("B", "C")
            data_ads = dut_ads.data[key_ads]
            data_ads = data_ads.calc_ft("B", "C")
            # data_simu    = dut_simu.data[key_simu]
            plt_gummel.add_data_set(
                np.asarray(abs(data_ngspice[data_ngspice["FREQ"] == 1e9]["V_B"])),
                np.asarray(abs(data_ngspice[data_ngspice["FREQ"] == 1e9]["I_C"])),
                label=r"$V_{{BC}} = \SI{" + str(vbc) + r"}{\volt}$",
            )
            plt_gummel.add_data_set(
                np.asarray(abs(data_ads[data_ads["FREQ"] == 1e9]["V_B"])),
                np.asarray(abs(data_ads[data_ads["FREQ"] == 1e9]["I_C"])),
            )
            plt_rey21.add_data_set(
                np.asarray(abs(data_ngspice[data_ngspice["FREQ"] == 1e9]["V_B"])),
                np.real(data_ngspice[data_ngspice["FREQ"] == 1e9]["Y_CB"]),
                label=r"$V_{{BC}} = \SI{" + str(vbc) + r"}{\volt}$",
            )
            plt_rey21.add_data_set(
                np.asarray(abs(data_ads[data_ads["FREQ"] == 1e9]["V_B"])),
                np.real(data_ads[data_ads["FREQ"] == 1e9]["Y_CB"]),
            )
            plt_imy21.add_data_set(
                np.asarray(abs(data_ngspice[data_ngspice["FREQ"] == 1e9]["V_B"])),
                np.imag(data_ngspice[data_ngspice["FREQ"] == 1e9]["Y_CB"]),
                label=r"$V_{{BC}} = \SI{" + str(vbc) + r"}{\volt}$",
            )
            plt_imy21.add_data_set(
                np.asarray(abs(data_ads[data_ads["FREQ"] == 1e9]["V_B"])),
                np.imag(data_ads[data_ads["FREQ"] == 1e9]["Y_CB"]),
            )
            plt_y11.add_data_set(
                np.asarray(abs(data_ngspice[data_ngspice["FREQ"] == 1e9]["V_B"])),
                np.asarray(abs(data_ngspice[data_ngspice["FREQ"] == 1e9]["Y_BB"])),
                label=r"$V_{{BC}} = \SI{" + str(vbc) + r"}{\volt}$",
            )
            plt_y11.add_data_set(
                np.asarray(abs(data_ads[data_ads["FREQ"] == 1e9]["V_B"])),
                np.asarray(abs(data_ads[data_ads["FREQ"] == 1e9]["Y_BB"])),
            )
            plt_ft.add_data_set(
                np.asarray(abs(data_ngspice[data_ngspice["FREQ"] == 1e9]["I_C"])),
                np.asarray(
                    abs(data_ngspice[data_ngspice["FREQ"] == 1e9][specifiers.TRANSIT_FREQUENCY])
                ),
                label=r"$V_{{BC}} = \SI{" + str(vbc) + r"}{\volt}$",
            )
            plt_ft.add_data_set(
                np.asarray(abs(data_ads[data_ads["FREQ"] == 1e9]["I_C"])),
                np.asarray(abs(data_ads[data_ads["FREQ"] == 1e9][specifiers.TRANSIT_FREQUENCY])),
            )

    if sim_ngspice:
        print("time ngspice:" + str(time_ngspice))
    if sim_ads:
        print("time ads    :" + str(time_ads))

    if show:
        for plt in plts[:-1]:
            plt.plot_pyqtgraph(show=False)

        plts[-1].plot_pyqtgraph(show=True)

    else:
        for plt in plts:
            plt.legend_location = "upper right outer"
            plt.save_tikz(
                test_path / "tmp",
                "ngspice_plot_" + plt.name,
                standalone=True,
                build=True,
                mark_repeat=3,
                width="0.4\\textwidth",
                height="0.3\\textwidth",
                fontsize="LARGE",
            )



""" singleton

Provides a meta class to ensure single instantiation.

"""
# DMT_core
# Copyright (C) from 2022  SemiMod
# Copyright (C) until 2021  Markus Müller, Mario Krattenmacher and Pascal Kuthe
# <https://gitlab.com/dmt-development/dmt-core>
#
# This file is part of DMT_core.
#
# DMT_core is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# DMT_core is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>


class Singleton(type):
    """Implements the Singleton design pattern. Classes that use this class as a metaclass can only be initiated once.

    If a new object of a singleton class is created, the already existing one is returned.
    Source: https://stackoverflow.com/questions/6760685/creating-a-singleton-in-python

    Here something else is added: If a already existing object is "created" again, it tries to transfer the given kwargs to the old object.

    Examples
    --------

    The simulation controller should only exist once because of possible blocking of resources. This class is used to see singleton in action:

    >>> sim_con1 = SimCon(n_core=4)
    >>> sim_con1.n_core
    4

    If now a second simulation controller is created:

    >>> sim_con2 = SimCon(n_core=8)
    >>> sim_con2.n_core
    8

    The new attribute is overwritten with the new value. As the change is made to the already existing instance, this also has now 8 cores:

    >>> sim_con1.n_core
    8

    """

    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(Singleton, cls).__call__(*args, **kwargs)
            obj = cls._instances[cls]
        else:
            obj = cls._instances[cls]
            if args:
                raise AttributeError(
                    "DMT: Can not set an value to an unknown attribute of the object.",
                    "Make sure, that all the wanted parameters are kwargs.",
                )

            # Try to set kwargs as parameters for the class.
            for attr, value in kwargs.items():
                if hasattr(obj, attr) and not callable(getattr(obj, attr)):
                    setattr(obj, attr, value)
                else:
                    raise AttributeError(
                        "DMT: Can not set an attribute which is not a member of the object"
                    )

        return obj
